@@ -6,7 +6,7 @@
  *                  Volker Sameske,  <sameske@de.ibm.com>
  *                  Holger Smolinski,<smolinsk@de.ibm.com>
  *                  Gerhard Tonn     <ton@de.ibm.com>
- * Copyright IBM Corp. 1999,2007
+ * Copyright IBM Corp. 1999, 2013
  */
 
 #include <sys/utsname.h>
@@ -14,7 +14,8 @@
 
 #include "zt_common.h"
 #include "dasdfmt.h"
-#include "vtoc.h" 
+#include "vtoc.h"
+#include "util_proc.h"
 
 /* Full tool name */
 static const char tool_name[] = "dasdfmt: zSeries DASD format program";
@@ -27,6 +28,7 @@ static int disk_disabled;
 static format_data_t format_params;
 char *prog_name;
 volatile sig_atomic_t program_interrupt_in_progress;
+int reqsize;
 
 /*
  * Print version information.
@@ -47,7 +49,8 @@ static void exit_usage(int exitcode)
 	       "	       [-l <volser>      | --label=<volser>]\n"
 	       "               [-b <blocksize>   | --blocksize=<blocksize>]\n"
 	       "               [-d <disk layout> | --disk_layout=<disk layout>]\n"
-	       "               <diskspec>\n\n",prog_name);
+	       "               [-r <cylinder>    | --requestsize=<cylinder>]\n"
+	       "               <device>\n\n",prog_name);
 
 	printf("       -t or --test     means testmode\n"
 	       "       -V or --version  means print version\n"
@@ -56,6 +59,8 @@ static void exit_usage(int exitcode)
 	       "       -P or --percentage means show a progress in percent\n"
 	       "       -m x or --hashmarks=x means show a hashmark every x "
 	       "cylinders\n"
+	       "       -r x or --requestsize=x means use x cylinders in one "
+	       "format step\n"
 	       "       -v means verbose mode\n"
 	       "       -F means don't check if the device is in use\n"
 	       "       -k means keep volume serial\n"
@@ -68,16 +73,7 @@ static void exit_usage(int exitcode)
 	       "       <disk layout> is either \n"
 	       "           'cdl' for compatible disk layout (default) or\n"
 	       "           'ldl' for linux disk layout\n"
-	       "       and <diskspec> is either\n"
-	       "           -f /dev/dasdX or --device=/dev/dasdX\n"
-	       "           if you do not use devfs\n"
-	       "         or\n"
-	       "           -f /dev/dasd/xxxx/device or "
-	       "--device=/dev/dasd/xxxx/device\n"
-	       "           and alternatively\n"
-	       "           -n xxxx or --devno=xxxx\n"
-	       "           in case you are using devfs.\n"
-	       "           xxxx is your hexadecimal device number.\n");
+	       "       <device> device node of the device to format\n");
 	exit(exitcode);
 }
 
@@ -140,32 +136,23 @@ static void program_interrupt_signal (int sig)
 
 /*
  * check given device name for blanks and some special characters
- * or create a devfs filename in case the devno was specified
  */
 static void get_device_name(dasdfmt_info_t *info, char *name, int argc, char * argv[])
 {
-	if (((info->node_specified + info->devno_specified) > 1) ||
-            ((info->node_specified + info->devno_specified) > 0 &&
-              info->device_id < argc)) 
-		ERRMSG_EXIT(EXIT_MISUSE,"%s: Device can only specified once! "
-			    "(%#04x or %s)\n", prog_name, info->devno, name);
+	struct util_proc_dev_entry dev_entry;
+	struct stat dev_stat;
 
-	if ((info->node_specified + info->devno_specified) == 0 &&
-            info->device_id >= argc) 
-		ERRMSG_EXIT(EXIT_MISUSE,"%s: No device specified!\n", 
+	if (info->node_specified && (info->device_id < argc))
+		ERRMSG_EXIT(EXIT_MISUSE,"%s: Device can only specified once!\n",
 			    prog_name);
 
-	if ((info->devno_specified) && 
-	    ((info->devno < 0x0000) || (info->devno > 0xffff)))
-		ERRMSG_EXIT(EXIT_MISUSE,"%s: Devno '%#04x' is not in range "
-			    "0x0000 - 0xFFFF!\n", prog_name, info->devno);
+	if (!info->node_specified && (info->device_id >= argc))
+		ERRMSG_EXIT(EXIT_MISUSE,"%s: No device specified!\n",
+			    prog_name);
 
-	if (info->devno_specified)
-		sprintf(info->devname, "/dev/dasd/%04x/device", info->devno);
-	else if (info->device_id < argc) {
+	if (info->device_id < argc) {
 		strcpy(info->devname, argv[info->device_id]);
-        }
-	else {
+        } else {
 		if ((strchr(name, ' ') != NULL)||(strchr(name, '#') != NULL)||
 		    (strchr(name, '[') != NULL)||(strchr(name, ']') != NULL)||
 		    (strchr(name, '!') != NULL)||(strchr(name, '>') != NULL)||
@@ -173,12 +160,33 @@ static void get_device_name(dasdfmt_info_t *info, char *name, int argc, char * a
 		    (strchr(name, ')') != NULL)||(strchr(name, ':') != NULL)||
 		    (strchr(name, '&') != NULL)||(strchr(name, ';') != NULL))
 			ERRMSG_EXIT(EXIT_MISUSE,"%s: Your filename contains "
-				    "blanks or special characters!\n", 
+				    "blanks or special characters!\n",
 				    prog_name);
 
 		strncpy(info->devname, name, PATH_MAX - 1);
 		info->devname[PATH_MAX - 1] = '\0';
 	}
+
+	if (stat(info->devname, &dev_stat) != 0)
+		ERRMSG_EXIT(EXIT_MISUSE,
+			    "%s: Could not get information for device node %s: %s\n",
+			    prog_name, info->devname, strerror(errno));
+
+	if (minor(dev_stat.st_rdev) & PARTN_MASK) {
+		ERRMSG_EXIT(EXIT_MISUSE,
+			    "%s: Unable to format partition %s. Please specify a device.\n",
+			    prog_name, info->devname);
+	}
+
+	if (util_proc_dev_get_entry(dev_stat.st_rdev, 1, &dev_entry) == 0) {
+		if (strncmp(dev_entry.name, "dasd", 4) != 0)
+			ERRMSG_EXIT(EXIT_MISUSE,
+				    "%s: Unsupported device type '%s'.\n",
+				    prog_name, dev_entry.name);
+	} else {
+		printf("%s WARNING: Unable to get driver name for device node %s",
+		       prog_name, info->devname);
+	}
 }
 
 
@@ -189,7 +197,6 @@ static void init_info(dasdfmt_info_t *info)
 {
 	info->devno             = 0x0;
 	info->usage_count       = 0;
-
 	info->testmode          = 0;
 	info->verbosity         = 0;
 	info->withoutprompt     = 0;
@@ -202,8 +209,8 @@ static void init_info(dasdfmt_info_t *info)
 	info->labelspec         = 0;
         info->cdl_format        = 0;
 	info->blksize_specified = 0;
+	info->reqsize_specified = 0;
 	info->node_specified    = 0;
-	info->devno_specified   = 0;
 	info->device_id         = 0;
 	info->keep_volser	= 0;
 }
@@ -245,18 +252,12 @@ static void check_disk(dasdfmt_info_t *info)
 	info->usage_count = dasd_info.open_count;
 	info->devno       = dasd_info.devno;
         if (strncmp(dasd_info.type, "ECKD",4) != 0) {
-		if (info->devno_specified) {
-			ERRMSG_EXIT(EXIT_FAILURE, 
-				    "%s: Unsupported disk type\n%x is not an "
-				    "ECKD disk!\n", prog_name, info->devno);
-		}
-		else {
-			ERRMSG_EXIT(EXIT_FAILURE, 
-				    "%s: Unsupported disk type\n%s is not an "
-				    "ECKD disk!\n", prog_name, info->devname);
-		}
-	}                                       
-}                                    
+		ERRMSG_EXIT(EXIT_FAILURE,
+			    "%s: Unsupported disk type\n%s is not an "
+			    "ECKD disk!\n", prog_name, info->devname);
+	}
+}
+
 
 
 /*
@@ -272,7 +273,7 @@ static int check_volser(char *s, int devno)
 		    ((s[i] >= 0x21)&&(s[i] <= 0x22)) ||  /*  !"        */
 		    ((s[i] >= 0x26)&&(s[i] <= 0x2f)) ||  /* &'()*+,-./ */
 		    ((s[i] >= 0x3a)&&(s[i] <= 0x3f)) ||  /*  :;<=>?    */
-		    ((s[i] >= 0x5b)&&(s[i] <= 0x60)))    /*  \]^_     */
+		    ((s[i] >= 0x5b)&&(s[i] <= 0x60)))    /*  \]^_`     */
 			s[i] = ' ';
 		s[i] = toupper(s[i]);
 	}
@@ -377,12 +378,8 @@ static void dasdfmt_print_info(dasdfmt_info_t *info, volume_label_t *vlabel,
 	       cylinders, heads, (cylinders * heads));
 
 	printf("\nI am going to format the device ");
-	if (info->devno_specified)
-		printf("%x in the following way:\n", info->devno);
-	else
-		printf("%s in the following way:\n", info->devname);
-
-	printf("   Device number of device : 0x%x\n",info->devno);
+	printf("%s in the following way:\n", info->devname);
+	printf("   Device number of device : 0x%x\n", info->devno);
 	printf("   Labelling device        : %s\n",
 	       (info->writenolabel)?"no":"yes");
 
@@ -398,7 +395,7 @@ static void dasdfmt_print_info(dasdfmt_info_t *info, volume_label_t *vlabel,
 	       (p->intensity & DASD_FMT_INT_COMPAT)?"yes":"no");
 	printf("   Blocksize               : %d\n", p->blksize);
 
-	if (info->testmode) 
+	if (info->testmode)
 		printf("Test mode active, omitting ioctl.\n");
 }
 
@@ -521,9 +518,8 @@ static void dasdfmt_write_labels(dasdfmt_info_t *info, volume_label_t *vlabel,
 			    "failed, only wrote %d bytes.\n", prog_name, rc);
 
 	/* write VTOC */
-	vtoc_init_format4_label(&f4, USABLE_PARTITIONS, geo.cylinders,
-				cylinders, heads, geo.sectors, blksize,
-				dasd_info.dev_type);
+	vtoc_init_format4_label(&f4, geo.cylinders, cylinders, heads,
+				geo.sectors, blksize, dasd_info.dev_type);
 
 	vtoc_init_format5_label(&f5);
 	vtoc_init_format7_label(&f7);
@@ -620,10 +616,12 @@ static void dasdfmt_format(dasdfmt_info_t *info, unsigned int cylinders,
 			   unsigned int heads, format_data_t *format_params)
 {
 	format_data_t format_step;
-	int j, cyl, tmp, p1, p2;
+	int j, cyl, tmp, p1, p2, hashcount = 0;
 	unsigned int k;
 
 	if (info->print_hashmarks) {
+		if (info->hashstep < reqsize)
+			info->hashstep = reqsize;
 		if ((info->hashstep < 1) || (info->hashstep > 1000)) {
 			printf("Hashmark increment is not in range <1,1000>, "
 			       "using the default.\n");
@@ -645,20 +643,21 @@ static void dasdfmt_format(dasdfmt_info_t *info, unsigned int cylinders,
 	while (1) {
 		p1 = -1; 
 		p2 =  0;
-
+		if (k + heads * reqsize >= format_params->stop_unit)
+			reqsize = 1;
 		format_step.start_unit = k;
-		format_step.stop_unit  = k + heads - 1;
+		format_step.stop_unit  = k + reqsize * heads - 1;
 
 		if (cyl == 1)
 			format_step.start_unit += 1;
-			
+
 		if (ioctl(filedes, BIODASDFMT, &format_step) != 0)
 			ERRMSG_EXIT(EXIT_FAILURE,"%s: (format cylinder) IOCTL "
 				    "BIODASDFMT failed. (%s)\n",
 				    prog_name, strerror(errno));
 
 		if (info->print_progressbar) {
-			printf("cyl %5d of %5d |", cyl, cylinders);
+			printf("cyl %7d of %7d |", cyl, cylinders);
 			p2 = p1;
 			p1 = cyl*100/cylinders;
 			if (p1 != p2)
@@ -669,26 +668,27 @@ static void dasdfmt_format(dasdfmt_info_t *info, unsigned int cylinders,
 					printf("#");
 				for (j=tmp+1; j<=50; j++) 
 					printf("-");
-				printf("| %3d%%  ", p1);
+				printf("|%3d%%", p1);
 			}
 			printf("\r");
 			fflush(stdout);
 		}
-		
+
 		if (info->print_hashmarks)
-			if (cyl % info->hashstep == 0) {
+			if ((cyl / info->hashstep - hashcount) != 0) {
 				printf("#");
 				fflush(stdout);
+				hashcount++;
 			}
 		if (info->print_percentage) {
-			printf("cyl %5i of %5i |  %3i%%\n", cyl, cylinders,
+			printf("cyl %7d of %7d |%3d%%\n", cyl, cylinders,
 			       cyl*100/cylinders);
 			fflush(stdout);
 		}
 
 		if (k % heads == 0) {
-			k += heads;
-			cyl++;
+			k += reqsize * heads;
+			cyl += reqsize;
 		}
 		else 
 			k += format_params->stop_unit % heads;
@@ -795,7 +795,7 @@ static void do_format_dasd(dasdfmt_info_t *info, format_data_t *p,
 			printf("You specified to write no labels to a"
 			       " volume with more then %u cylinders.\n"
 			       "Cylinders above this limit will not be"
-			       " accessable as a linux partition!\n"
+			       " accessible as a linux partition!\n"
 			       "Type \"yes\" to continue, no will leave"
 			       " the disk untouched: ", LV_COMPAT_CYL);
 			if (fgets(inp_buffer, sizeof(inp_buffer), stdin) == NULL)
@@ -847,7 +847,7 @@ static void do_format_dasd(dasdfmt_info_t *info, format_data_t *p,
 		if (!((info->withoutprompt)&&(info->verbosity<1)))
 			printf("Formatting the device. This may take a "
 			       "while (get yourself a coffee).\n");
-		
+
 		dasdfmt_prepare_and_format(info, cylinders, heads, p);
 
 		printf("Finished formatting the device.\n");
@@ -875,8 +875,8 @@ int main(int argc,char *argv[])
 	char str[ERR_LENGTH];
 	char buf[7];
 
-	char *devno_param_str   = NULL;
 	char *blksize_param_str = NULL;
+	char *reqsize_param_str = NULL;
 	char *hashstep_str      = NULL;
 
 	int rc, index;
@@ -999,12 +999,10 @@ int main(int argc,char *argv[])
 			blksize_param_str=optarg;
 			info.blksize_specified=1;
 			break;
-			
-		case 'n' :
-			devno_param_str=optarg;
-			info.devno_specified=1;
+		case 'r':
+			reqsize_param_str = optarg;
+			info.reqsize_specified = 1;
 			break;
-		
 		case 'f' :
 			strncpy(dev_filename, optarg, PATH_MAX);
 			info.node_specified=1;
@@ -1028,12 +1026,17 @@ int main(int argc,char *argv[])
 	CHECK_SPEC_MAX_ONCE(info.labelspec, "label");
 	CHECK_SPEC_MAX_ONCE(info.writenolabel, "omit-label-writing flag");
 
-	if (info.devno_specified)
-		PARSE_PARAM_INTO(info.devno, devno_param_str, 16,
-				 "device number");
 	if (info.blksize_specified)
 		PARSE_PARAM_INTO(format_params.blksize,blksize_param_str,10,
 				 "blocksize");
+	if (info.reqsize_specified) {
+		PARSE_PARAM_INTO(reqsize, reqsize_param_str, 10, "requestsize");
+		if (reqsize < 1 || reqsize > 255)
+			ERRMSG_EXIT(EXIT_FAILURE,
+				    "invalid requestsize %d specified\n",
+				    reqsize);
+	} else
+		reqsize = DEFAULT_REQUESTSIZE;
 	if (info.print_hashmarks)
 		PARSE_PARAM_INTO(info.hashstep, hashstep_str,10,"hashstep");
 
