@@ -1,12 +1,17 @@
 /*
  * File...........: s390-tools/fdasd/fdasd.c
- * Author(s)......: Volker Sameske <sameske@de.ibm.com>
- *                  Horst Hummel   <horst.hummel@de.ibm.com>
- *                  Gerhard Tonn   <ton@de.ibm.com>
- * Copyright IBM Corp. 2001,2011
+ * Author(s)......: Volker Sameske   <sameske@de.ibm.com>
+ *                  Horst Hummel     <horst.hummel@de.ibm.com>
+ *                  Gerhard Tonn     <ton@de.ibm.com>
+ *                  Stefan Weinhuber <wein@de.ibm.com>
+ * Copyright IBM Corp. 2001,2012
  */
 
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
 #include <getopt.h>
+#include <stdio.h>
 
 #include "zt_common.h"
 #include "vtoc.h"
@@ -286,18 +291,25 @@ read_char(char *mesg)
  * Print question string an enforce y/n answer.
  */
 static int
-yes_no(char *question_str) 
+yes_no(char *question_str)
 {
-	char answer;
-	int rc;
+	char *answer;
+	size_t size;
+	ssize_t bytes_read;
+
+	size = 0;
+	answer = NULL;
 	while (1) {
 		printf("%s (y/n): ", question_str);
-		rc = scanf("%c", &answer);
-		if (rc != 1)
+		bytes_read = getline(&answer, &size, stdin);
+		if (bytes_read < 0)
 			return -1;
-		if (answer == 'y') return 0;
-		if (answer == 'n') return 1;
+		if (answer[0] == 'y')
+			return 0;
+		if (answer[0] == 'n')
+			return 1;
 	}
+	free(answer);
 }
 
 
@@ -353,7 +365,8 @@ fdasd_usage (void)
 		                          "using information\n"
                 "                         found in CONFIGFILE\n"
 		"-i, --volser             Print volume serial\n"
-		"-p, --table              Print partition table\n");
+		"-p, --table              Print partition table\n"
+		"-f, --force              Force fdasd to work on non DASD devices\n");
 }
 
 
@@ -461,6 +474,44 @@ fdasd_initialize_anchor (fdasd_anchor_t *anc)
 	anc->formatted_cylinders = 0;
 }
 
+static void fdasd_parse_force_options(fdasd_anchor_t *anc, char *optarg)
+{
+	unsigned int devtype, blksize;
+	char err_str[ERROR_STRING_SIZE];
+	int rc;
+
+	if (optarg) {
+		rc = sscanf(optarg, "%x,%d", &devtype, &blksize);
+		if (rc != 2) {
+			snprintf(err_str, ERROR_STRING_SIZE,
+				 "Force parameter '%s' could not be parsed.\n",
+				 optarg);
+			fdasd_error(anc, parser_failed, err_str);
+		}
+		if (devtype == DASD_3390_TYPE || devtype == DASD_3380_TYPE ||
+		    devtype == DASD_9345_TYPE)
+			anc->dev_type = devtype;
+		else {
+			snprintf(err_str, ERROR_STRING_SIZE,
+				 "Force parameter '%x' is not a supported"
+				 " device type.\n", devtype);
+			fdasd_error(anc, parser_failed, err_str);
+		}
+		if (blksize == 4096 || blksize == 2048 ||
+		    blksize == 1024 || blksize ==  512)
+			anc->blksize = blksize;
+		else {
+			snprintf(err_str, ERROR_STRING_SIZE,
+				 "Force parameter '%d' is not a supported"
+				 " block size.\n", blksize);
+			fdasd_error(anc, parser_failed, err_str);
+		}
+	} else {
+		/* force option was used without the optional argument */
+		anc->dev_type = DASD_3390_TYPE;
+		anc->blksize = 4096;
+	}
+}
 
 /*
  * parses the command line options
@@ -514,6 +565,10 @@ fdasd_parse_options (fdasd_anchor_t *anc, struct fdasd_options *options,
 		case 'k':
 			anc->keep_volser++;
 			break;
+		case 'f':
+			anc->force_virtual++;
+			fdasd_parse_force_options(anc, optarg);
+			break;
 		case -1:
 			/* End of options string - start of devices list */
 			break;
@@ -803,7 +858,7 @@ fdasd_verify_device (fdasd_anchor_t *anc, char *name)
  *
  * Note: 
  *  - 'version' and 'help' are priority options. 
- *       All other paramters are ignored in that case.
+ *       All other parameters are ignored in that case.
  *  - 'silent' and 'verbose' are allways allowed in any
  *       combination.
  * 
@@ -1032,49 +1087,14 @@ fdasd_list_partition_table (fdasd_anchor_t *anc)
  * get volser from vtoc
  */
 static int
-fdasd_get_volser(fdasd_anchor_t *anc, char *devname, char *volser)
+fdasd_get_volser(fdasd_anchor_t *anc, char *volser)
 {
-	dasd_information_t dasd_info;
-	int blksize;
-	int fd;
 	volume_label_t vlabel;
-	char err_str[ERROR_STRING_SIZE];
 
-	if ((fd = open(devname, O_RDONLY)) == -1) {
-		snprintf(err_str, ERROR_STRING_SIZE,
-			"Could not open device '%s' "
-			"in read-only mode!\n", devname);
-		fdasd_error(anc, unable_to_open_disk, err_str);
-	}     
-
-	if (ioctl(fd, BIODASDINFO, &dasd_info) != 0) {
-		close(fd);
-		fdasd_error(anc, unable_to_ioctl, 
-			    "Could not retrieve disk information.");
-	}
-
-	if (ioctl(fd, BLKSSZGET, &blksize) != 0) {
-		close(fd);
-	        fdasd_error(anc, unable_to_ioctl,
-			    "Could not retrieve blocksize information.");
-	}
-
-	close(fd);
-
-	if ((strncmp(dasd_info.type, "ECKD", 4) == 0) &&
-	    (!dasd_info.FBA_layout)) {
-		/* OS/390 and zOS compatible disk layout */
-		vtoc_read_volume_label(options.device,
-				       dasd_info.label_block * blksize,
-				       &vlabel);
-		vtoc_volume_label_get_volser(&vlabel, volser);
-		return 0;
-        }
-	else {
-		return -1;
-	}
+	vtoc_read_volume_label(options.device, anc->label_pos, &vlabel);
+	vtoc_volume_label_get_volser(&vlabel, volser);
+	return 0;
 }
-	
 
 /*
  * call IOCTL to re-read the partition table
@@ -1289,8 +1309,8 @@ static void fdasd_recreate_vtoc_unconditional(fdasd_anchor_t *anc)
 	partition_info_t *part_info = anc->first;
 	int i;
 
-	vtoc_init_format4_label(anc->f4, USABLE_PARTITIONS,
-				geo.cylinders, anc->formatted_cylinders,
+	vtoc_init_format4_label(anc->f4, geo.cylinders,
+				anc->formatted_cylinders,
 				geo.heads, geo.sectors,
 				anc->blksize, anc->dev_type);
 
@@ -1375,8 +1395,7 @@ fdasd_reuse_vtoc(fdasd_anchor_t *anc)
 
 	if (!anc->silent) printf("re-creating VTOC... ");
 
-	vtoc_init_format4_label(&f4, USABLE_PARTITIONS,
-				geo.cylinders, anc->formatted_cylinders,
+	vtoc_init_format4_label(&f4, geo.cylinders, anc->formatted_cylinders,
 				geo.heads, geo.sectors,
 				anc->blksize, anc->dev_type);
 
@@ -1401,12 +1420,10 @@ fdasd_reuse_vtoc(fdasd_anchor_t *anc)
 		}
 
 		if (anc->formatted_cylinders > LV_COMPAT_CYL)
-			vtoc_init_format8_label(anc->vlabel->volid,
-						anc->blksize,
+			vtoc_init_format8_label(anc->blksize,
 						&part_info->f1->DS1EXT1, &f1);
 		else
-			vtoc_init_format1_label(anc->vlabel->volid,
-						anc->blksize,
+			vtoc_init_format1_label(anc->blksize,
 						&part_info->f1->DS1EXT1, &f1);
 
 
@@ -1554,7 +1571,7 @@ fdasd_init_volume_label(fdasd_anchor_t *anc)
 	vtoc_volume_label_set_label(vlabel, "VOL1");
 
 	if (anc->keep_volser) {
-		if(fdasd_get_volser(anc, options.device, volser) == 0)
+		if (fdasd_get_volser(anc, volser) == 0)
 			vtoc_volume_label_set_volser(vlabel, volser);
 		else
 			fdasd_error(anc, volser_not_found, options.device);
@@ -1680,8 +1697,8 @@ fdasd_process_invalid_vtoc(fdasd_anchor_t *anc)
 	anc->formatted_cylinders = anc->hw_cylinders;
 	anc->fspace_trk = anc->formatted_cylinders * geo.heads
 		- FIRST_USABLE_TRK;
-	vtoc_init_format4_label(anc->f4, USABLE_PARTITIONS,
-				geo.cylinders, anc->formatted_cylinders,
+	vtoc_init_format4_label(anc->f4, geo.cylinders,
+				anc->formatted_cylinders,
 				geo.heads, geo.sectors,
 				anc->blksize, anc->dev_type);
 
@@ -1855,8 +1872,8 @@ fdasd_invalid_vtoc_pointer(fdasd_anchor_t *anc)
 	anc->formatted_cylinders = anc->hw_cylinders;
 	anc->fspace_trk = anc->formatted_cylinders * geo.heads
 		- FIRST_USABLE_TRK;
-	vtoc_init_format4_label(anc->f4, USABLE_PARTITIONS,
-				geo.cylinders, anc->formatted_cylinders,
+	vtoc_init_format4_label(anc->f4, geo.cylinders,
+				anc->formatted_cylinders,
 				geo.heads, geo.sectors,
 				anc->blksize, anc->dev_type);
 
@@ -1956,8 +1973,8 @@ fdasd_check_volume (fdasd_anchor_t *anc)
 
 		fdasd_init_volume_label(anc);
 
-		vtoc_init_format4_label(anc->f4, USABLE_PARTITIONS,
-					geo.cylinders, anc->formatted_cylinders,
+		vtoc_init_format4_label(anc->f4, geo.cylinders,
+					anc->formatted_cylinders,
 					geo.heads, geo.sectors,
 					anc->blksize, anc->dev_type);
 
@@ -2034,17 +2051,77 @@ fdasd_check_disk_access (fdasd_anchor_t *anc)
         close(fd);
 }     
 
+/*
+ * The following two functions match those in the DASD ECKD device driver.
+ * They are used to compute how many records of a given size can be stored
+ * in one track.
+ */
+static unsigned int ceil_quot(unsigned int d1, unsigned int d2)
+{
+	return (d1 + (d2 - 1)) / d2;
+}
+
+/* kl: key length, dl: data length */
+static unsigned int recs_per_track(unsigned short dev_type, unsigned int kl,
+				   unsigned int dl)
+{
+	int dn, kn;
+
+	switch (dev_type) {
+	case DASD_3380_TYPE:
+		if (kl)
+			return 1499 / (15 + 7 + ceil_quot(kl + 12, 32) +
+				       ceil_quot(dl + 12, 32));
+		else
+			return 1499 / (15 + ceil_quot(dl + 12, 32));
+	case DASD_3390_TYPE:
+		dn = ceil_quot(dl + 6, 232) + 1;
+		if (kl) {
+			kn = ceil_quot(kl + 6, 232) + 1;
+			return 1729 / (10 + 9 + ceil_quot(kl + 6 * kn, 34) +
+				       9 + ceil_quot(dl + 6 * dn, 34));
+		} else
+			return 1729 / (10 + 9 + ceil_quot(dl + 6 * dn, 34));
+	case DASD_9345_TYPE:
+		dn = ceil_quot(dl + 6, 232) + 1;
+		if (kl) {
+			kn = ceil_quot(kl + 6, 232) + 1;
+			return 1420 / (18 + 7 + ceil_quot(kl + 6 * kn, 34) +
+				       ceil_quot(dl + 6 * dn, 34));
+		} else
+			return 1420 / (18 + 7 + ceil_quot(dl + 6 * dn, 34));
+	}
+	return 0;
+}
+
+/*
+ * Verify that number of tracks (heads) per cylinder and number of
+ * sectors per track match the expected values for a given device type
+ * and block size.
+ * Returns 1 for a valid match and 0 otherwise.
+ */
+static int fdasd_verify_geometry(unsigned short dev_type, int blksize,
+				 struct hd_geometry *geometry)
+{
+	unsigned int expected_sectors;
+	if (geometry->heads != 15)
+		return 0;
+	expected_sectors = recs_per_track(dev_type, 0, blksize);
+	if (geometry->sectors == expected_sectors)
+		return 1;
+	return 0;
+}
 
 /*
  * reads dasd geometry data
  */
-static void
-fdasd_get_geometry (fdasd_anchor_t *anc) 
+static void fdasd_get_geometry (fdasd_anchor_t *anc)
 {
         int fd, blksize = 0;
 	dasd_information_t dasd_info;
 	char err_str[ERROR_STRING_SIZE];
 	struct dasd_eckd_characteristics *characteristics;
+	unsigned long long size_in_bytes;
 
 	if ((fd = open(options.device,O_RDONLY)) < 0) {
 		snprintf(err_str, ERROR_STRING_SIZE,
@@ -2052,75 +2129,137 @@ fdasd_get_geometry (fdasd_anchor_t *anc)
 			"in read-only mode!\n", options.device);
 		fdasd_error(anc, unable_to_open_disk, err_str);
 	}
+	if (ioctl(fd, BLKGETSIZE64, &size_in_bytes) != 0) {
+		close(fd);
+		fdasd_error(anc, unable_to_ioctl,
+			    "Could not retrieve disk size.");
+	}
+	/*
+	 * If anc->force_virtual is set, we do no real geometry detection.
+	 * anc->dev_type and anc->blksize have already been set via command
+	 * line parameter, and the rest of the geometry is now computed from
+	 * these values.
+	 */
+	if (anc->force_virtual) {
+		geo.heads = 15;
+		geo.sectors = recs_per_track(anc->dev_type, 0, anc->blksize);
+		anc->hw_cylinders = size_in_bytes /
+			(anc->blksize * geo.heads * geo.sectors);
+		if (anc->hw_cylinders < LV_COMPAT_CYL)
+			geo.cylinders = anc->hw_cylinders;
+		else
+			geo.cylinders = LV_COMPAT_CYL;
+		geo.start = 0;
+		anc->label_pos = 2 * anc->blksize;
+		anc->devno = 0;
+		close(fd);
+		if (anc->verbose)
+			printf("The force option is active. "
+			       "The following geometry will be used:\n"
+			       "device type %x, block size %d, cylinders %d,"
+			       " heads %d, sectors %d \n",
+			       anc->dev_type, anc->blksize, anc->hw_cylinders,
+			       geo.heads, geo.sectors);
+		return;
+	}
 
 	if (ioctl(fd, HDIO_GETGEO, &geo) != 0) {
-	        close(fd);
-		fdasd_error(anc, unable_to_ioctl, 
+		close(fd);
+		fdasd_error(anc, unable_to_ioctl,
 			    "Could not retrieve disk geometry information.");
 	}
-
 	if (ioctl(fd, BLKSSZGET, &blksize) != 0) {
-	        close(fd);
-	        fdasd_error(anc, unable_to_ioctl,
+		close(fd);
+		fdasd_error(anc, unable_to_ioctl,
 			    "Could not retrieve blocksize information.");
 	}
-
 	/* get disk type */
 	if (ioctl(fd, BIODASDINFO, &dasd_info) != 0) {
-	        close(fd);
-		fdasd_error(anc, unable_to_ioctl, 
-			    "Could not retrieve disk information.");
-	}
+		if (anc->verbose)
+			printf("BIODASDINFO ioctl failed,"
+			       " use disk geometry only.\n");
+		/* verify that the geometry matches a 3390 DASD */
+		if (!fdasd_verify_geometry(DASD_3390_TYPE, blksize, &geo)) {
+			close(fd);
+			fdasd_error(anc, wrong_disk_type,
+				    "Disk geometry does not match a DASD device"
+				    " of type 3390.");
+		}
+		anc->dev_type = DASD_3390_TYPE;
+		anc->blksize = blksize;
+		anc->hw_cylinders =
+			size_in_bytes /	(blksize * geo.heads * geo.sectors);
+		/* The VOL1 label on a CDL formatted ECKD DASD is in block 2
+		 * It will be verified later, if this position actually holds a
+		 * valid label record.
+		 */
+		anc->label_pos = 2 * blksize;
+		/* A devno 0 is actually a valid devno, which could exist
+		 * in the system. Since we use this number only to create
+		 * a default volume serial, there is no serious conflict.
+		 */
+		anc->devno = 0;
+		if (anc->verbose)
+			printf("The following device geometry will be used:\n"
+			       "device type %x, block size %d, cylinders %d,"
+			       " heads %d, sectors %d \n",
+			       anc->dev_type, anc->blksize, anc->hw_cylinders,
+			       geo.heads, geo.sectors);
 
-	characteristics =
-		(struct dasd_eckd_characteristics *) &dasd_info.characteristics;
-	if (characteristics->no_cyl == LV_COMPAT_CYL &&
-	    characteristics->long_no_cyl)
-		anc->hw_cylinders = characteristics->long_no_cyl;
-	else
-		anc->hw_cylinders = characteristics->no_cyl;
+
+	} else {
+		characteristics = (struct dasd_eckd_characteristics *)
+			&dasd_info.characteristics;
+		if (characteristics->no_cyl == LV_COMPAT_CYL &&
+		    characteristics->long_no_cyl)
+			anc->hw_cylinders = characteristics->long_no_cyl;
+		else
+			anc->hw_cylinders = characteristics->no_cyl;
+
+		if (strncmp(dasd_info.type, "ECKD", 4) != 0) {
+			snprintf(err_str, ERROR_STRING_SIZE,
+				 "%s is not an ECKD disk! This disk type "
+				 "is not supported!", options.device);
+			fdasd_error(anc,wrong_disk_type, err_str);
+		}
+
+		if (anc->verbose) printf("disk type check     : ok\n");
+
+		if (dasd_info.FBA_layout != 0) {
+			snprintf(err_str, ERROR_STRING_SIZE,
+				 "%s is not formatted with z/OS compatible "
+				 "disk layout!", options.device);
+			fdasd_error(anc, wrong_disk_format, err_str);
+		}
+
+		if (anc->verbose) printf("disk layout check   : ok\n");
+
+		if (dasd_info.open_count > 1) {
+			if (anc->auto_partition) {
+				snprintf(err_str, ERROR_STRING_SIZE,
+					 "DASD '%s' is in use. Unmount it first!",
+					 options.device);
+				fdasd_error(anc, disk_in_use, err_str);
+			} else {
+				printf("\nWARNING: Your DASD '%s' is in use.\n"
+				       "         If you proceed, you can "
+				       "heavily damage your system.\n"
+				       "         If possible exit all"
+				       " applications using this disk\n"
+				       "         and/or unmount it.\n\n",
+				       options.device);
+			}
+		}
+
+		if (anc->verbose) printf("usage count check   : ok\n");
+
+		anc->dev_type   = dasd_info.dev_type;
+		anc->blksize    = blksize;
+		anc->label_pos  = dasd_info.label_block * blksize;
+		anc->devno      = dasd_info.devno;
+	}
 
 	close(fd);
-
-	if (strncmp(dasd_info.type, "ECKD", 4) != 0) {
-		snprintf(err_str, ERROR_STRING_SIZE, 
-			"%s is not an ECKD disk! This disk type "
-			"is not supported!", options.device);
-		fdasd_error(anc,wrong_disk_type, err_str);
-	}
-
-	if (anc->verbose) printf("disk type check     : ok\n");
-
-        if (dasd_info.FBA_layout != 0) {
-                snprintf(err_str, ERROR_STRING_SIZE, 
-			"%s is not formatted with z/OS compatible "
-			"disk layout!", options.device);
-                fdasd_error(anc, wrong_disk_format, err_str);
-        }      
-
-	if (anc->verbose) printf("disk layout check   : ok\n");
-
-	if (dasd_info.open_count > 1) {
-		if (anc->auto_partition) {
-			snprintf(err_str, ERROR_STRING_SIZE, 
-				"DASD '%s' is in use. Unmount it first!",
-				options.device);
-			fdasd_error(anc, disk_in_use, err_str);
-		} else {	
-			printf("\nWARNING: Your DASD '%s' is in use.\n"
-			       "         If you proceed, you can heavily "
-			       "damage your system.\n         If possible exit"
-			       " all applications using this disk\n         "
-			       "and/or unmount it.\n\n", options.device);
-		}
-	}
-
-	if (anc->verbose) printf("usage count check   : ok\n");
-
-	anc->dev_type   = dasd_info.dev_type;
-	anc->blksize    = blksize;
-	anc->label_pos  = dasd_info.label_block * blksize;
-	anc->devno      = dasd_info.devno;
 }
 
 
@@ -2490,11 +2629,9 @@ fdasd_add_partition (fdasd_anchor_t *anc)
 	        return;
 
 	if (anc->formatted_cylinders > LV_COMPAT_CYL) {
-		vtoc_init_format8_label(anc->vlabel->volid, anc->blksize, &ext,
-					part_info->f1);
+		vtoc_init_format8_label(anc->blksize, &ext, part_info->f1);
 	} else
-		vtoc_init_format1_label(anc->vlabel->volid, anc->blksize, &ext,
-					 part_info->f1);
+		vtoc_init_format1_label(anc->blksize, &ext, part_info->f1);
 
 	fdasd_enqueue_new_partition(anc);
 	anc->used_partitions += 1;
@@ -2583,8 +2720,8 @@ fdasd_auto_partition(fdasd_anchor_t *anc)
 	fdasd_init_volume_label(anc);
 
 	if (anc->verbose) printf("initializing labels...\n");
-	vtoc_init_format4_label(anc->f4, USABLE_PARTITIONS,
-				geo.cylinders, anc->formatted_cylinders,
+	vtoc_init_format4_label(anc->f4, geo.cylinders,
+				anc->formatted_cylinders,
 				geo.heads, geo.sectors,
 				anc->blksize, anc->dev_type);
 
@@ -2606,11 +2743,9 @@ fdasd_auto_partition(fdasd_anchor_t *anc)
 	vtoc_set_extent(&ext, 0x01, 0x00, &llimit, &ulimit);
 
 	if (anc->formatted_cylinders > LV_COMPAT_CYL) {
-		vtoc_init_format8_label(anc->vlabel->volid, anc->blksize, &ext,
-					part_info->f1);
+		vtoc_init_format8_label(anc->blksize, &ext, part_info->f1);
 	} else
-		vtoc_init_format1_label(anc->vlabel->volid, anc->blksize, &ext,
-					part_info->f1);
+		vtoc_init_format1_label(anc->blksize, &ext, part_info->f1);
         anc->fspace_trk      = 0;
 	anc->used_partitions = 1;
 
@@ -2630,7 +2765,6 @@ fdasd_auto_partition(fdasd_anchor_t *anc)
 static void
 fdasd_auto_partition_conffile(fdasd_anchor_t *anc)
 {
-	volume_label_t *vlabel = anc->vlabel;
 	partition_info_t *part_info = anc->first;
 	cchh_t llimit,ulimit;
 	unsigned long start, stop;
@@ -2641,8 +2775,8 @@ fdasd_auto_partition_conffile(fdasd_anchor_t *anc)
 	fdasd_init_volume_label(anc);
 
 	if (anc->verbose) printf("initializing labels...\n");
-	vtoc_init_format4_label(anc->f4, USABLE_PARTITIONS,
-				geo.cylinders, anc->formatted_cylinders,
+	vtoc_init_format4_label(anc->f4, geo.cylinders,
+				anc->formatted_cylinders,
 				geo.heads, geo.sectors,
 				anc->blksize, anc->dev_type);
 
@@ -2674,11 +2808,11 @@ fdasd_auto_partition_conffile(fdasd_anchor_t *anc)
 				0x00, &llimit, &ulimit);
 
 		if (anc->formatted_cylinders > LV_COMPAT_CYL) {
-			vtoc_init_format8_label(vlabel->volid, anc->blksize,
-						&ext, part_info->f1);
+			vtoc_init_format8_label(anc->blksize, &ext,
+						part_info->f1);
 		} else
-			vtoc_init_format1_label(vlabel->volid, anc->blksize,
-						&ext, part_info->f1);
+			vtoc_init_format1_label(anc->blksize, &ext,
+						part_info->f1);
 		anc->used_partitions += 1;
 
 		get_addr_of_highest_f1_f8_label(anc, &hf1);
