@@ -2,7 +2,7 @@
  * File...........: s390-tools/fdasd/fdasd.h
  * Author(s)......: Volker Sameske <sameske@de.ibm.com>
  *                  Horst Hummel   <Horst.Hummel@de.ibm.com>
- * Copyright IBM Corp. 2001,2007
+ * Copyright IBM Corp. 2001,2012
  */
 
 #ifndef FDASD_H
@@ -166,11 +166,12 @@ static struct option fdasd_long_options[] = {
 	{ "table",       no_argument,       NULL, 'p'},
 	{ "volser",      no_argument,       NULL, 'i'},
 	{ "keep_volser", no_argument,       NULL, 'k'},
+	{ "force",       optional_argument, NULL, 'f'},
 	{ 0,             0,                 0,    0  }
 };
 
 /* Command line option abbreviations */
-static const char option_string[] = "vasrl:c:hpik";
+static const char option_string[] = "vasrl:c:hpikf::";
 
 struct fdasd_options {
 	char *device;
@@ -211,6 +212,7 @@ typedef struct fdasd_anchor {
 	int print_table;
 	int print_volser;
 	int keep_volser;
+	int force_virtual;
 	int big_disk;
 	int silent;
 	int verbose;
