@@ -29,7 +29,7 @@ static int l_row_is_marked(struct table *t, struct table_row *row)
 {
 	struct table_mark_key *key;
 
-	list_iterate(key, &t->mark_key_list, list) {
+	util_list_iterate(&t->mark_key_list, key) {
 		if (strcmp(row->entries[0].str, key->str) == 0)
 			return 1;
 	}
@@ -45,7 +45,7 @@ static void l_mark_key_add(struct table *t, char *str)
 
 	key = ht_zalloc(sizeof(*key));
 	strncpy(key->str, str, sizeof(key->str));
-	list_add_end(&key->list, &t->mark_key_list);
+	util_list_add_tail(&t->mark_key_list, key);
 	t->mark_keys_cnt++;
 }
 
@@ -54,11 +54,11 @@ static void l_mark_key_add(struct table *t, char *str)
  */
 static void l_mark_key_remove(struct table *t, char *str)
 {
-	struct table_mark_key *key;
+	struct table_mark_key *key, *tmp;
 
-	list_iterate(key, &t->mark_key_list, list) {
+	util_list_iterate_safe(&t->mark_key_list, key, tmp) {
 		if (strcmp(str, key->str) == 0) {
-			list_del(&key->list);
+			util_list_remove(&t->mark_key_list, key);
 			ht_free(key);
 			t->mark_keys_cnt--;
 			return;
@@ -74,10 +74,10 @@ void table_row_mark_del_all(struct table *t)
 	struct table_mark_key *key, *tmp;
 	struct table_row *row;
 
-	list_iterate(row, &t->row_list, list)
+	util_list_iterate(&t->row_list, row)
 		row->marked = 0;
-	list_iterate_safe(key, &t->mark_key_list, list, tmp) {
-		list_del(&key->list);
+	util_list_iterate_safe(&t->mark_key_list, key, tmp) {
+		util_list_remove(&t->mark_key_list, key);
 		ht_free(key);
 	}
 	t->mark_keys_cnt = 0;
@@ -108,7 +108,7 @@ void table_row_mark_toggle_by_key(struct table *t, const char *str)
 {
 	struct table_row *row;
 
-	list_iterate(row, &t->row_list, list) {
+	util_list_iterate(&t->row_list, row) {
 		if (strcmp(str, row->entries[0].str) == 0)
 			table_row_mark_toggle(t, row);
 	}
@@ -148,7 +148,6 @@ struct table_row *table_row_alloc(struct table *t)
 	table_row = ht_zalloc(sizeof(*table_row));
 	table_row->entries = ht_zalloc(sizeof(*table_row->entries) *
 					  t->col_cnt);
-	list_init(&table_row->list);
 	return table_row;
 }
 
@@ -169,8 +168,8 @@ struct table *table_new(int extra_rows, int sorted, int first_bold,
 {
 	struct table *t = ht_zalloc(sizeof(*t));
 
-	list_init(&t->row_list);
-	list_init(&t->mark_key_list);
+	util_list_init(&t->row_list, struct table_row, list);
+	util_list_init(&t->mark_key_list, struct table_mark_key, list);
 	t->row_cnt_marked = 0;
 	if (with_units)
 		t->row_cnt_extra = extra_rows + L_ROWS_EXTRA + 1;
@@ -254,8 +253,8 @@ void table_row_del_all(struct table *t)
 {
 	struct table_row *row, *tmp;
 
-	list_iterate_safe(row, &t->row_list, list, tmp) {
-		list_del(&row->list);
+	util_list_iterate_safe(&t->row_list, row, tmp) {
+		util_list_remove(&t->row_list, row);
 		table_row_free(row);
 	}
 	l_row_last_init(t);
@@ -400,7 +399,7 @@ static void l_row_last_calc(struct table *t)
 	struct table_row *row;
 
 	l_row_last_init(t);
-	list_iterate(row, &t->row_list, list) {
+	util_list_iterate(&t->row_list, row) {
 		if (t->mode_hide_unmarked && !row->marked)
 			continue;
 		l_row_last_agg(t, row);
@@ -426,14 +425,17 @@ void table_row_add(struct table *t, struct table_row *row)
 
 	l_row_format(t, row);
 
-	if (list_is_empty(&t->row_list) || !t->attr_sorted_table) {
-		list_add_end(&row->list, &t->row_list);
+	if (util_list_is_empty(&t->row_list) || !t->attr_sorted_table) {
+		util_list_add_tail(&t->row_list, row);
 	} else {
-		list_iterate(tmp, &t->row_list, list) {
+		util_list_iterate(&t->row_list, tmp) {
 			if (l_row_less_than(t, tmp, row))
 				break;
 		}
-		list_add_end(&row->list, &tmp->list);
+		if (tmp)
+			util_list_add_prev(&t->row_list, row, tmp);
+		else
+			util_list_add_tail(&t->row_list, row);
 	}
 	if (l_row_is_marked(t, row)) {
 		row->marked = 1;
@@ -453,45 +455,25 @@ void table_rebuild(struct table *t)
 
 	table_col_iterate(t, col, i)
 		l_col_max_width_init(t, col);
-	list_iterate(row, &t->row_list, list)
+	util_list_iterate(&t->row_list, row)
 		l_row_format(t, row);
 	l_row_format(t, t->row_last);
 }
 
 /*
- * Sort table (TODO: Use better sorting algorithm)
+ * Compare callback for linked list sorting (ordering: large to small)
+ */
+static int l_row_cmp_fn(void *a, void *b, void *data)
+{
+	return l_row_less_than(data, a, b) ? 1 : -1;
+}
+
+/*
+ * Sort table (ordering: large to small)
  */
 static void l_table_sort(struct table *t)
 {
-	struct table_row *row_min;
-	struct table_row *row;
-	struct table_row *tmp;
-	struct list list;
-
-	list_init(&list);
-
-	/*
-	 * Sort row list into temp list
-	 */
-	while (!list_is_empty(&t->row_list)) {
-		row_min = NULL;
-
-		list_iterate(row, &t->row_list, list) {
-			if (row_min == NULL)
-				row_min = row;
-			else if (l_row_less_than(t, row, row_min))
-				row_min = row;
-		}
-		list_del(&row_min->list);
-		list_add(&row_min->list, &list);
-	}
-	/*
-	 * Copy temp list to original list
-	 */
-	list_iterate_safe(row, &list, list, tmp) {
-		list_del(&row->list);
-		list_add_end(&row->list, &t->row_list);
-	}
+	util_list_sort(&t->row_list, l_row_cmp_fn, t);
 }
 
 /*
@@ -585,7 +567,7 @@ static struct table_row *l_selected_row(struct table *t)
 	struct table_row *row;
 	int row_nr = 0;
 
-	list_iterate(row, &t->row_list, list) {
+	util_list_iterate(&t->row_list, row) {
 		if (t->mode_hide_unmarked && !row->marked)
 			continue;
 		if (row_nr == t->row_nr_select)
@@ -904,7 +886,7 @@ static void l_table_print_curses(struct table *t)
 	l_status_update(t);
 	l_headline_print(t);
 	l_unitline_print(t);
-	list_iterate(row, &t->row_list, list) {
+	util_list_iterate(&t->row_list, row) {
 		if (t->mode_hide_unmarked && !row->marked)
 			continue;
 		if (row_nr < t->row_nr_begin) {
@@ -958,7 +940,7 @@ static void l_table_print_all(struct table *t)
 
 	l_headline_print(t);
 	l_unitline_print(t);
-	list_iterate(row, &t->row_list, list) {
+	util_list_iterate(&t->row_list, row) {
 		l_row_print(t, row);
 		hyptop_print_nl();
 	}
