@@ -26,7 +26,8 @@
 
 #include "error.h"
 #include "misc.h"
-#include "proc.h"
+#include "util_proc.h"
+#include "install.h"
 
 /* from linux/fs.h */
 #define FIBMAP			_IO(0x00,1)
@@ -75,9 +76,9 @@ disk_determine_dasd_type(struct disk_info *data,
 		data->type = disk_type_diag;
 	else if (strncmp(dasd_info.type, "ECKD",4) == 0) {
 		if (dasd_info.FBA_layout)
-			data->type = disk_type_eckd_classic;
+			data->type = disk_type_eckd_ldl;
 		else
-			data->type = disk_type_eckd_compatible;
+			data->type = disk_type_eckd_cdl;
 	} else {
 		error_reason("Unknown DASD type");
 		return -1;
@@ -89,8 +90,101 @@ disk_determine_dasd_type(struct disk_info *data,
 int
 disk_is_eckd(disk_type_t type)
 {
-	return (type == disk_type_eckd_classic ||
-		type == disk_type_eckd_compatible);
+	return (type == disk_type_eckd_ldl ||
+		type == disk_type_eckd_cdl);
+}
+
+static int
+read_block_by_offset(int fd, int blksize, uint64_t offset, char *buffer)
+{
+
+	if (lseek(fd, offset, SEEK_SET) == -1) {
+		/* Seek error. */
+		error_text("Error: Could not seek to %llu: %s!\n",
+			   (unsigned long long) offset);
+		return -1;
+	}
+
+	misc_read(fd, buffer, blksize);
+
+	return 0;
+}
+
+int
+determine_virtblk_type(struct disk_info *data, struct stat *stats)
+{
+	char *device;
+	char *buffer;
+	int fd, rc, shift, sb;
+
+	rc = 0;
+	buffer = (char *) malloc(data->phy_block_size);
+	if (!buffer)
+		return -1;
+
+	/*
+	 * The geo.start value reported for virtblk devices is based on a
+	 * 512 byte blocksize.
+	 * For DASD devices it is based on the real (most likely 4k) blocksize
+	 * and the DASD device driver reports a shifted value.
+	 * For virtblk devices we need to shift the value manually according to
+	 * the physical blocksize of the device.
+	 */
+	shift = 0;
+	for (sb = 512; sb < data->phy_block_size; sb = sb << 1)
+		shift++;
+
+	if (data->geo.heads == 15) {
+		/* assume DASD */
+		data->partnum = stats->st_rdev & DASD_PARTN_MASK;
+		data->device = stats->st_rdev & ~DASD_PARTN_MASK;
+
+		rc = misc_temp_dev(data->device, 1, &device);
+		if (rc)
+			goto out_err;
+
+		fd = open(device, O_RDONLY);
+
+		/* read 3rd record, containing the volume label */
+		read_block_by_offset(fd, data->phy_block_size,
+				     2 * data->phy_block_size, buffer);
+		misc_ebcdic_to_ascii((unsigned char *) buffer,
+				     (unsigned char *) buffer + 4);
+		/* determine dasd type by label */
+		if (!strncmp(buffer, "VOL1", 4)) {
+			data->type = disk_type_eckd_cdl;
+			data->geo.start >>= shift;
+		} else if (!strncmp(buffer, "LNX1", 4)) {
+			data->type = disk_type_eckd_ldl;
+			data->geo.start >>= shift;
+		} else if (!strncmp(buffer, "CMS1", 4)) {
+			data->type = disk_type_eckd_ldl;
+			data->geo.start >>= shift;
+		} else {
+			/* DASD label was not found,
+			 * type has to be specified by hand
+			 */
+			error_text("Failed to read DASD label, "
+				   "please specify type manually");
+			rc = -1;
+		}
+		close(fd);
+		misc_free_temp_dev(device);
+	} else if (data->geo.heads == 16) {
+		error_text("Assume disk_type FBA for virtblk device, "
+			   "but please specify type manually");
+		data->geo.start >>= shift;
+		disk_print_geo(data);
+		rc = -1;
+	} else {
+		data->type = disk_type_scsi;
+		data->partnum = stats->st_rdev & SCSI_PARTN_MASK;
+		data->device = stats->st_rdev & ~SCSI_PARTN_MASK;
+	}
+
+out_err:
+	free(buffer);
+	return rc;
 }
 
 int
@@ -99,8 +193,8 @@ disk_get_info(const char* device, struct job_target_data* target,
 {
 	struct stat stats;
 	struct stat script_stats;
-	struct proc_part_entry part_entry;
-	struct proc_dev_entry dev_entry;
+	struct util_proc_part_entry part_entry;
+	struct util_proc_dev_entry dev_entry;
 	struct dasd_information dasd_info;
 	struct disk_info *data;
 	int fd;
@@ -133,9 +227,9 @@ disk_get_info(const char* device, struct job_target_data* target,
 	}
 	memset((void *) data, 0, sizeof(struct disk_info));
 	/* Try to get device driver name */
-	if (proc_dev_get_entry(stats.st_rdev, 1, &dev_entry) == 0) {
+	if (util_proc_dev_get_entry(stats.st_rdev, 1, &dev_entry) == 0) {
 		data->drv_name = misc_strdup(dev_entry.name);
-		proc_dev_free_entry(&dev_entry);
+		util_proc_dev_free_entry(&dev_entry);
 	} else {
 		fprintf(stderr, "Warning: Could not determine driver name for "
 			"major %d from /proc/devices\n", major(stats.st_rdev));
@@ -153,8 +247,15 @@ disk_get_info(const char* device, struct job_target_data* target,
 		data->source = source_script;
 		/* Run targetbase script */
 		strcpy(ppn_cmd, script_file);
-		strcat(ppn_cmd, " ");
-		strcat(ppn_cmd, target->bootmap_dir);
+		if (target->bootmap_dir == NULL) {
+			/* happens in case of partition dump */
+			snprintf(ppn_cmd, sizeof(ppn_cmd), "%s %d:%d",
+				 script_file, major(stats.st_rdev),
+				 minor(stats.st_rdev));
+		} else {
+			strcat(ppn_cmd, " ");
+			strcat(ppn_cmd, target->bootmap_dir);
+		}
 		printf("Run %s\n", ppn_cmd);
 		fh = popen(ppn_cmd, "r");
 		if (fh == NULL) {
@@ -223,24 +324,6 @@ disk_get_info(const char* device, struct job_target_data* target,
 	if ((data->source == source_user) || (data->source == source_script)) {
 		data->devno = -1;
 		data->phy_block_size = target->targetblocksize;
-	} else {
-		/* Get DASD information */
-		if (ioctl(fd, BIODASDINFO, &dasd_info))
-			data->devno = -1;
-		else
-			data->devno = dasd_info.devno;
-		/* Get physical block size */
-		if (ioctl(fd, BLKSSZGET, &data->phy_block_size)) {
-			error_reason("Could not get blocksize");
-			goto out_close;
-		}
-	}
-	/* Get size of device in sectors (512 byte) */
-	if (ioctl(fd, BLKGETSIZE, &devsize)) {
-		error_reason("Could not get device size");
-		goto out_close;
-	}
-	if ((data->source == source_user) || (data->source == source_script)) {
 		data->type = target->targettype;
 		data->partnum = 0;
 		/* Get file information */
@@ -248,6 +331,7 @@ disk_get_info(const char* device, struct job_target_data* target,
 		    == 2) {
 			data->device = makedev(majnum, minnum);
 			data->targetbase = defined_as_device;
+			data->partnum = minor(stats.st_rdev) - minnum;
 		}
 		else {
 			if (stat(target->targetbase, &stats)) {
@@ -267,10 +351,15 @@ disk_get_info(const char* device, struct job_target_data* target,
 		}
 		goto type_determined;
 	}
+	if (ioctl(fd, BLKSSZGET, &data->phy_block_size)) {
+		error_reason("Could not get blocksize");
+		goto out_close;
+	}
 	/* Determine disk type */
 	if (!data->drv_name) {
 		/* Driver name cannot be read */
-		if (data->devno == -1) {
+		if (ioctl(fd, BIODASDINFO, &dasd_info)) {
+			data->devno = -1;
 			if (data->geo.start) {
 				/* SCSI partition */
 				data->type = disk_type_scsi;
@@ -284,6 +373,7 @@ disk_get_info(const char* device, struct job_target_data* target,
 			}
 		} else {
 			/* DASD */
+			data->devno = dasd_info.devno;
 			if (disk_determine_dasd_type(data, dasd_info))
 				goto out_close;
 			data->partnum = stats.st_rdev & DASD_PARTN_MASK;
@@ -291,25 +381,34 @@ disk_get_info(const char* device, struct job_target_data* target,
 		}
 	} else if (strcmp(data->drv_name, "dasd") == 0) {
 		/* Driver name is 'dasd' */
-		if (data->devno == -1) {
+		if (ioctl(fd, BIODASDINFO, &dasd_info)) {
 			error_reason("Could not determine DASD type");
 			goto out_close;
 		}
+		data->devno = dasd_info.devno;
 		if (disk_determine_dasd_type(data, dasd_info))
 			goto out_close;
 		data->partnum = stats.st_rdev & DASD_PARTN_MASK;
 		data->device = stats.st_rdev & ~DASD_PARTN_MASK;
-	} else if (strcmp(data->drv_name, "sd") == 0 ||
-		   strcmp(data->drv_name, "virtblk") == 0) {
-		/* Driver name is 'sd' or 'virtblk' */
-		if (data->devno != -1) {
-			error_reason("Unsupported device driver '%s' "
-				     "for disk type DASD", data->drv_name);
-			goto out_close;
-		}
+	} else if (strcmp(data->drv_name, "sd") == 0) {
+		/* Driver name is 'sd' */
+		data->devno = -1;
 		data->type = disk_type_scsi;
 		data->partnum = stats.st_rdev & SCSI_PARTN_MASK;
 		data->device = stats.st_rdev & ~SCSI_PARTN_MASK;
+
+	} else if (strcmp(data->drv_name, "virtblk") == 0) {
+
+		if (ioctl(fd, HDIO_GETGEO, &data->geo) != 0)
+			perror("Could not retrieve disk geometry information.");
+		if (ioctl(fd, BLKSSZGET, &data->phy_block_size) != 0)
+			perror("Could not retrieve blocksize information.");
+
+		if (determine_virtblk_type(data, &stats)) {
+			error_reason("Virtblk device type not clearly "
+				     "determined.");
+			goto out_close;
+		}
 	} else {
 		/* Driver name is unknown */
 		error_reason("Unsupported device driver '%s'", data->drv_name);
@@ -317,6 +416,12 @@ disk_get_info(const char* device, struct job_target_data* target,
 	}
 
 type_determined:
+	/* Get size of device in sectors (512 byte) */
+	if (ioctl(fd, BLKGETSIZE, &devsize)) {
+		error_reason("Could not get device size");
+		goto out_close;
+	}
+
 	/* Check for valid CHS geometry data. */
 	if (disk_is_eckd(data->type) && (data->geo.cylinders == 0 ||
 	    data->geo.heads == 0 || data->geo.sectors == 0)) {
@@ -330,9 +435,9 @@ type_determined:
 	if (data->partnum != 0)
 		data->partition = stats.st_rdev;
 	/* Try to get device name */
-	if (proc_part_get_entry(data->device, &part_entry) == 0) {
+	if (util_proc_part_get_entry(data->device, &part_entry) == 0) {
 		data->name = misc_strdup(part_entry.name);
-		proc_part_free_entry(&part_entry);
+		util_proc_part_free_entry(&part_entry);
 		if (data->name == NULL)
 			goto out_close;
 	}
@@ -350,6 +455,34 @@ out_close:
 
 }
 
+int
+disk_is_tape(const char* device)
+{
+	int fd, rc = 0;
+
+	/* Check for tape */
+	fd = open(device, O_RDWR);
+	if (fd == -1)
+		return 0;
+	if (rewind_tape(fd) == 0)
+		rc = 1;
+	close(fd);
+	return rc;
+}
+
+int
+disk_is_scsi(const char* device, struct job_target_data* target)
+{
+	struct disk_info* info;
+	int rc = 0;
+
+	if (disk_get_info(device, target, &info) == -1)
+		return 0;
+	if (info->type == disk_type_scsi)
+		rc = 1;
+	disk_free_info(info);
+	return rc;
+}
 
 int
 disk_get_info_from_file(const char* filename, struct job_target_data* target,
@@ -417,14 +550,23 @@ disk_free_info(struct disk_info* info)
  * blocknumber in the variable pointed to by PHYSICAL. Return non-zero
  * otherwise. */
 int
-disk_get_blocknum(int fd, blocknum_t logical, blocknum_t* physical,
-		  struct disk_info* info)
+disk_get_blocknum(int fd, int fd_is_basedisk, blocknum_t logical,
+		  blocknum_t* physical, struct disk_info* info)
 {
 	struct statfs buf;
 	blocknum_t phy_per_fs;
 	int mapped;
 	int subblock;
 
+	/* No file system: partition or raw disk */
+	if (info->fs_block_size == -1) {
+		if (fd_is_basedisk)
+			*physical = logical;
+		else
+			*physical = logical + info->geo.start;
+		return 0;
+	}
+
 	/* Get file system type */
 	if (fstatfs(fd, &buf)) {
 		error_reason(strerror(errno));
@@ -500,8 +642,8 @@ disk_blockptr_from_blocknum(disk_blockptr_t* ptr, blocknum_t blocknum,
 		ptr->linear.size = info->phy_block_size;
 		ptr->linear.blockct = 0;
 		break;
-	case disk_type_eckd_classic:
-	case disk_type_eckd_compatible:
+	case disk_type_eckd_ldl:
+	case disk_type_eckd_cdl:
 		if (blocknum == 0) {
 			/* Special case: zero blocknum will be expanded to
 			 * size * (blockct+1) bytes of zeroes. */
@@ -528,9 +670,10 @@ disk_blockptr_from_blocknum(disk_blockptr_t* ptr, blocknum_t blocknum,
  * bytes are written. INFO provides information about the disk layout. Upon
  * success, return 0 and store the pointer to the resulting disk block to BLOCK
  * (if BLOCK is not NULL). Return non-zero otherwise. */
-int
-disk_write_block_aligned(int fd, const void* data, size_t bytecount,
-			 disk_blockptr_t* block, struct disk_info* info)
+static int
+disk_write_block_aligned_base(int fd, int is_base_disk, const void* data,
+			      size_t bytecount, disk_blockptr_t* block,
+			      struct disk_info* info)
 {
 	blocknum_t current_block;
 	blocknum_t blocknum;
@@ -560,13 +703,21 @@ disk_write_block_aligned(int fd, const void* data, size_t bytecount,
 		return -1;
 	if (block != NULL) {
 		/* Store block pointer */
-		if (disk_get_blocknum(fd, current_block, &blocknum, info))
+		if (disk_get_blocknum(fd, is_base_disk, current_block,
+				      &blocknum, info))
 			return -1;
 		disk_blockptr_from_blocknum(block, blocknum, info);
 	}
 	return 0;
 }
 
+int
+disk_write_block_aligned(int fd, const void* data, size_t bytecount,
+			 disk_blockptr_t* block, struct disk_info* info)
+{
+	return disk_write_block_aligned_base(fd, 0, data, bytecount, block,
+					     info);
+}
 
 /* Write BYTECOUNT bytes from memory at location BUFFER to the file identified
  * by file descriptor FD and return the list of pointers to the disk blocks
@@ -575,8 +726,8 @@ disk_write_block_aligned(int fd, const void* data, size_t bytecount,
  * otherwise. Note that the data is written to a file position which is aligned
  * on a block size boundary. */
 blocknum_t
-disk_write_block_buffer(int fd, const void* buffer, size_t bytecount,
-			disk_blockptr_t** blocklist,
+disk_write_block_buffer(int fd, int fd_is_basedisk, const void* buffer,
+			size_t bytecount, disk_blockptr_t** blocklist,
 			struct disk_info* info)
 {
 	disk_blockptr_t* list;
@@ -599,7 +750,8 @@ disk_write_block_buffer(int fd, const void* buffer, size_t bytecount,
 		chunk_size = bytecount - written;
 		if (chunk_size > (size_t) info->phy_block_size)
 			chunk_size = info->phy_block_size;
-		rc = disk_write_block_aligned(fd, VOID_ADD(buffer, written),
+		rc = disk_write_block_aligned_base(fd, fd_is_basedisk,
+						   VOID_ADD(buffer, written),
 				chunk_size,
 				&list[i], info);
 		if (rc) {
@@ -631,9 +783,9 @@ disk_get_type_name(disk_type_t type)
 		return "FBA disk layout";
 	case disk_type_diag:
 		return "DIAG disk layout";
-	case disk_type_eckd_classic:
+	case disk_type_eckd_ldl:
 		return "ECKD/linux disk layout";
-	case disk_type_eckd_compatible:
+	case disk_type_eckd_cdl:
 		return "ECKD/compatible disk layout";
 	default:
 		return "Unknown disk type";
@@ -645,8 +797,8 @@ disk_get_type_name(disk_type_t type)
 int
 disk_is_large_volume(struct disk_info* info)
 {
-	return (info->type == disk_type_eckd_classic ||
-	        info->type == disk_type_eckd_compatible) &&
+	return (info->type == disk_type_eckd_ldl ||
+		info->type == disk_type_eckd_cdl) &&
 		info->geo.cylinders == 0xfffe;
 }
 
@@ -682,8 +834,8 @@ disk_print_info(struct disk_info* info)
 	}
 	if (((info->type == disk_type_fba) ||
 	     (info->type == disk_type_diag) ||
-	     (info->type == disk_type_eckd_classic) ||
-	     (info->type == disk_type_eckd_compatible)) &&
+	     (info->type == disk_type_eckd_ldl) ||
+	     (info->type == disk_type_eckd_cdl)) &&
 	     (info->source == source_auto)) {
 		printf("  DASD device number..............: %04x\n",
 		       info->devno);
@@ -721,6 +873,16 @@ disk_print_info(struct disk_info* info)
 		printf("  *) Data provided by script.\n");
 }
 
+/* Print textual representation of geo structure. */
+void
+disk_print_geo(struct disk_info *data)
+{
+		printf("  geo.heads.........:%u\n", data->geo.heads);
+		printf("  geo.sectors.......:%u\n", data->geo.sectors);
+		printf("  geo.cylinders.....:%u\n", data->geo.cylinders);
+		printf("  geo.start.........:%lu\n", data->geo.start);
+		printf("  blksize...........:%u\n", data->phy_block_size);
+}
 
 /* Check whether a block is a zero block which identifies a hole in a file.
  * Return non-zero if BLOCK is a zero block, 0 otherwise. */
@@ -731,8 +893,8 @@ disk_is_zero_block(disk_blockptr_t* block, struct disk_info* info)
 	case disk_type_scsi:
 	case disk_type_fba:
 		return block->linear.block == 0;
-	case disk_type_eckd_classic:
-	case disk_type_eckd_compatible:
+	case disk_type_eckd_ldl:
+	case disk_type_eckd_cdl:
 		return (block->chs.cyl == 0) && (block->chs.head == 0) &&
 		       (block->chs.sec == 0);
 	default:
@@ -773,8 +935,8 @@ can_merge_blocks(disk_blockptr_t* first, disk_blockptr_t* second,
 		    second->linear.block)
 			return 1;
 		break;
-	case disk_type_eckd_classic:
-	case disk_type_eckd_compatible:
+	case disk_type_eckd_ldl:
+	case disk_type_eckd_cdl:
 		/* Check link count limits */
 		if (((int) first->chs.blockct) +
 		    ((int) second->chs.blockct) + 1 > max_count)
@@ -804,8 +966,8 @@ merge_blocks(disk_blockptr_t* first, disk_blockptr_t* second,
 	case disk_type_fba:
 		first->linear.blockct += second->linear.blockct + 1;
 		break;
-	case disk_type_eckd_classic:
-	case disk_type_eckd_compatible:
+	case disk_type_eckd_ldl:
+	case disk_type_eckd_cdl:
 		first->chs.blockct += second->chs.blockct + 1;
 		break;
 	case disk_type_diag:
@@ -883,7 +1045,7 @@ disk_get_blocklist_from_file(const char* filename, disk_blockptr_t** blocklist,
 	memset((void *) list, 0, sizeof(disk_blockptr_t) * count);
 	/* Build list */
 	for (i=0; i < count; i++) {
-		if (disk_get_blocknum(fd, i, &blocknum, info)) {
+		if (disk_get_blocknum(fd, 0, i, &blocknum, info)) {
 			free(list);
 			close(fd);
 			return 0;
