@@ -30,6 +30,16 @@
 #include "error.h"
 #include "misc.h"
 
+#define ECKD_CDL_DUMP_REC 3 /* Start record (0 based) for stage 2 and 1b */
+/* CDL record 2 = 148 byte, independent from bs, bootinfo is last structure */
+#define CDL_BOOTINFO_ADDR 112
+/* ADDRESS for bootinfo structure to be BIOS compatible in MBR */
+#define DEFAULT_BOOTINFO_ADDR 408
+
+static inline unsigned long blk_cnt(int size, struct disk_info *info)
+{
+	return (size + info->phy_block_size - 1) / info->phy_block_size;
+}
 
 /* Types of SCSI disk layouts */
 enum scsi_layout {
@@ -39,6 +49,9 @@ enum scsi_layout {
 	scsi_layout_unknown
 };
 
+/* From linux/fs.h */
+#define BLKFLSBUF	_IO(0x12, 97)
+
 
 /* Determine SCSI disk layout from the specified BOOTBLOCK. */
 static enum scsi_layout
@@ -57,19 +70,25 @@ get_scsi_layout(unsigned char* bootblock)
 
 #define DISK_LAYOUT_ID 0x00000001
 
+static int
+overwrite_partition_start(int fd, struct disk_info* info, int mv_dump_magic);
+
 /* Create an IPL master boot record data structure for SCSI MBRs in memory
  * at location BUFFER. TABLE contains a pointer to the program table. INFO
  * provides information about the disk. */
 static int
 update_scsi_mbr(void* bootblock, disk_blockptr_t* table,
-		struct disk_info* info)
+		struct disk_info* info, disk_blockptr_t* scsi_dump_sb_blockptr)
 {
 	struct scsi_mbr {
 		uint8_t		magic[4];
 		uint32_t	version_id;
 		uint8_t		reserved[8];
 		uint8_t		program_table_pointer[16];
+		uint8_t		reserved2[0x50];
+		struct boot_info boot_info;
 	}  __attribute__ ((packed))* mbr;
+	struct scsi_dump_param param;
 	void* buffer;
 
 	switch (get_scsi_layout(bootblock)) {
@@ -98,6 +117,13 @@ update_scsi_mbr(void* bootblock, disk_blockptr_t* table,
 	memcpy(&mbr->magic, ZIPL_MAGIC, ZIPL_MAGIC_SIZE);
 	mbr->version_id = DISK_LAYOUT_ID;
 	bootmap_store_blockptr(&mbr->program_table_pointer, table, info);
+	if (scsi_dump_sb_blockptr->linear.block != 0) {
+		/* Write dump boot_info */
+		param.block = scsi_dump_sb_blockptr->linear.block *
+			      scsi_dump_sb_blockptr->linear.size;
+		boot_get_dump_info(&mbr->boot_info, BOOT_INFO_DEV_TYPE_SCSI,
+				   &param);
+	}
 	return 0;
 }
 
@@ -107,7 +133,8 @@ update_scsi_mbr(void* bootblock, disk_blockptr_t* table,
  * to the disk block containing the program table. INFO provides
  * information about the disk type. Return 0 on success, non-zero otherwise. */
 static int
-install_scsi(int fd, disk_blockptr_t* program_table, struct disk_info* info)
+install_scsi(int fd, disk_blockptr_t* program_table, struct disk_info* info,
+	     disk_blockptr_t* scsi_dump_sb_blockptr)
 {
 	unsigned char* bootblock;
 	int rc;
@@ -116,9 +143,7 @@ install_scsi(int fd, disk_blockptr_t* program_table, struct disk_info* info)
 	if (bootblock == NULL)
 		return -1;
 	/* Read bootblock */
-	if (lseek(fd, 0, SEEK_SET) != 0) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
+	if (misc_seek(fd, 0)) {
 		free(bootblock);
 		return -1;
 	}
@@ -129,7 +154,8 @@ install_scsi(int fd, disk_blockptr_t* program_table, struct disk_info* info)
 		return rc;
 	}
 	/* Put zIPL data into MBR */
-	rc = update_scsi_mbr(bootblock, program_table, info);
+	rc = update_scsi_mbr(bootblock, program_table, info,
+			     scsi_dump_sb_blockptr);
 	if (rc) {
 		free(bootblock);
 		return -1;
@@ -137,9 +163,7 @@ install_scsi(int fd, disk_blockptr_t* program_table, struct disk_info* info)
 	/* Write MBR back to disk */
 	if (verbose)
 		printf("Writing SCSI master boot record.\n");
-	if (lseek(fd, 0, SEEK_SET) != 0) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
+	if (misc_seek(fd, 0)) {
 		free(bootblock);
 		return -1;
 	}
@@ -153,142 +177,116 @@ install_scsi(int fd, disk_blockptr_t* program_table, struct disk_info* info)
 
 /* Install bootloader for initial program load from an FBA type disk. */
 static int
-install_fba(int fd, disk_blockptr_t* program_table,
-	    disk_blockptr_t* stage2_list, blocknum_t stage2_count,
-	    struct disk_info* info)
+install_fba(int fd, disk_blockptr_t *program_table,
+	    disk_blockptr_t *stage1b_list, blocknum_t stage1b_count,
+	    struct disk_info *info)
 {
 	struct boot_fba_stage0 stage0;
-	int rc;
 
-	/* Initialize stage 0 data */
-	rc = boot_init_fba_stage0(&stage0, stage2_list, stage2_count);
-	if (rc == 0) {
-		/* Store program table pointer in stage 0 loader */
-		bootmap_store_blockptr(&stage0.param1, program_table, info);
-		/* Install stage 0 */
-		rc = DRY_RUN_FUNC(misc_write(
-			fd, &stage0, sizeof(struct boot_fba_stage0)));
-	}
-	if (rc)
-		error_text("Could not write boot loader");
-	return rc;
+	/* Install stage 0 and store program table pointer */
+	if (boot_init_fba_stage0(&stage0, stage1b_list, stage1b_count))
+		return -1;
+
+	boot_get_ipl_info(&stage0.boot_info,  BOOT_INFO_DEV_TYPE_FBA,
+			  program_table, info);
+
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage0, sizeof(stage0), 0)))
+		return -1;
+
+	return 0;
 }
 
+/* Install stage1b bootloader for ECKD type disk */
+int
+install_eckd_stage1b(int fd, disk_blockptr_t **stage1b_list,
+		     blocknum_t *stage1b_count, disk_blockptr_t *stage2_list,
+		     blocknum_t stage2_count, struct disk_info *info)
+{
+	struct boot_eckd_stage1b *stage1b;
+	int stage1b_size, rc = -1;
+
+	*stage1b_list = NULL;
+	*stage1b_count = 0;
+	stage1b_size = ROUNDUP(sizeof(*stage1b), info->phy_block_size);
+	stage1b = misc_malloc(stage1b_size);
+	if (stage1b == NULL)
+		goto out;
+	memset(stage1b, 0, stage1b_size);
+	if (boot_init_eckd_stage1b(stage1b, stage2_list, stage2_count))
+		goto out_free_stage1b;
+	*stage1b_count = disk_write_block_buffer(fd, 1, stage1b, stage1b_size,
+						 stage1b_list, info);
+	if (*stage1b_count == 0)
+		goto out_free_stage1b;
+	rc = 0;
+out_free_stage1b:
+	free(stage1b);
+out:
+	return rc;
+}
 
 /* Install bootloader for initial program load from an ECKD type disk with
- * Linux/390 classic disk layout. */
+ * Linux Disk Layout. */
 static int
-install_eckd_classic(int fd, disk_blockptr_t* program_table,
-		     disk_blockptr_t* stage2_list, blocknum_t stage2_count,
-		     struct disk_info* info)
+install_eckd_ldl(int fd, disk_blockptr_t *program_table,
+		 disk_blockptr_t *stage1b_list, blocknum_t stage1b_count,
+		 struct disk_info *info)
 {
-	struct boot_eckd_stage0 stage0;
-	struct boot_eckd_classic_stage1 stage1;
-	int rc;
+	struct boot_eckd_ldl_stage0 stage0;
+	struct boot_eckd_stage1 stage1;
 
-	/* Prepare stage 0 */
-	boot_init_eckd_classic_stage0(&stage0);
 	/* Install stage 0 */
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage0,
-		sizeof(struct boot_eckd_stage0)));
-	if (rc) {
-		error_text("Could not write boot loader");
-		return rc;
-	}
-	/* Prepare stage 1 */
-	rc = boot_init_eckd_classic_stage1(&stage1, stage2_list, stage2_count);
-	if (rc)
-		return rc;
-	/* Store program table pointer in stage 1 loader */
-	bootmap_store_blockptr(&stage1.param1, program_table, info);
-	/* Install stage 1 */
-	if (lseek(fd, info->phy_block_size, SEEK_SET) !=
-	    info->phy_block_size) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
+	boot_init_eckd_ldl_stage0(&stage0);
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage0, sizeof(stage0), 0)))
 		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage1,
-		sizeof(struct boot_eckd_classic_stage1)));
-	if (rc) {
-		error_text("Could not write boot loader");
-		return rc;
-	}
-	return rc;
+	/* Install stage 1 and store program table pointer */
+	if (boot_init_eckd_stage1(&stage1, stage1b_list, stage1b_count))
+		return -1;
+
+	boot_get_ipl_info(&stage1.boot_info, BOOT_INFO_DEV_TYPE_ECKD,
+			  program_table, info);
+
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage1, sizeof(stage1),
+				     sizeof(stage0))))
+		return -1;
+
+	return 0;
 }
 
-
-
 /* Install bootloader for initial program load from an ECKD type disk with
  * OS/390 compatible disk layout. */
 static int
-install_eckd_compatible(int fd, disk_blockptr_t* program_table,
-			struct disk_info* info, struct job_data* job)
+install_eckd_cdl(int fd, disk_blockptr_t *program_table,
+		 disk_blockptr_t *stage1b_list, blocknum_t stage1b_count,
+		 struct disk_info *info)
 {
-	struct boot_eckd_stage0 stage0;
-	struct boot_eckd_compatible_stage1 stage1;
-	void* buffer;
-	size_t size;
-	int rc;
+	struct boot_eckd_cdl_stage0 stage0;
+	struct boot_eckd_stage1 stage1;
 
-	/* Prepare stage 0 */
-	boot_init_eckd_compatible_stage0(&stage0);
 	/* Install stage 0 */
-	if (lseek(fd, 4, SEEK_SET) != 4) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
+	boot_init_eckd_cdl_stage0(&stage0);
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage0, sizeof(stage0), 4)))
 		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage0,
-		sizeof(struct boot_eckd_stage0)));
-	if (rc) {
-		error_text("Could not write boot loader");
+	/* Install stage 1 and store program table pointer */
+	if (boot_init_eckd_stage1(&stage1, stage1b_list, stage1b_count))
 		return -1;
-	}
-	/* Prepare stage 1 */
-	rc = boot_init_eckd_compatible_stage1(&stage1, info);
-	if (rc)
-		return rc;
-	/* Store program table pointer in stage 1 loader */
-	bootmap_store_blockptr(&stage1.param1, program_table, info);
-	/* Install stage 1 */
-	if (lseek(fd, 4 + info->phy_block_size, SEEK_SET) !=
-	    4 + info->phy_block_size) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
+
+	boot_get_ipl_info(&stage1.boot_info, BOOT_INFO_DEV_TYPE_ECKD,
+			  program_table, info);
+
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage1, sizeof(stage1),
+				     4 + info->phy_block_size)))
 		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage1,
-		sizeof(struct boot_eckd_compatible_stage1)));
-	if (rc) {
-		error_text("Could not write boot loader");
-		return -1;
-	}
-	/* Install stage 2 */
-	if (lseek(fd, 3 * info->phy_block_size, SEEK_SET) !=
-	    3 * info->phy_block_size) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = boot_get_eckd_stage2(&buffer, &size, job);
-	if (rc)
-		return -1;
-	rc = DRY_RUN_FUNC(misc_write(fd, buffer, size));
-	if (rc)
-		error_text("Could not write boot loader");
-	free(buffer);
-	return rc;
+	return 0;
 }
 
-
 int
-install_bootloader(const char* device, disk_blockptr_t *program_table,
-		   disk_blockptr_t* stage2_list, blocknum_t stage2_count,
-		   struct disk_info* info, struct job_data* job)
+install_bootloader(const char *device, disk_blockptr_t *program_table,
+		   disk_blockptr_t *scsi_dump_sb_blockptr,
+		   disk_blockptr_t *stage1b_list, blocknum_t stage1b_count,
+		   struct disk_info *info, struct job_data *job)
 {
-	int fd;
-	int rc;
+	int fd, rc;
 
 	/* Inform user about what we're up to */
 	printf("Preparing boot device: ");
@@ -311,22 +309,34 @@ install_bootloader(const char* device, disk_blockptr_t *program_table,
 			   device);
 		return -1;
 	}
+	/* Ensure that potential cache inconsistencies between disk and
+	 * partition are resolved by flushing the corresponding buffers. */
+	if (!dry_run) {
+		if (ioctl(fd, BLKFLSBUF)) {
+			fprintf(stderr, "Warning: Could not flush disk "
+				"caches.\n");
+		}
+	}
 	/* Call disk specific install functions */
 	rc = -1;
 	switch (info->type) {
 	case disk_type_scsi:
-		rc = install_scsi(fd, program_table, info);
+		rc = install_scsi(fd, program_table, info,
+				  scsi_dump_sb_blockptr);
+		if (rc == 0 && job->id == job_dump_partition)
+			rc = overwrite_partition_start(fd, info, 0);
 		break;
 	case disk_type_fba:
-		rc = install_fba(fd, program_table, stage2_list, stage2_count,
-				 info);
+		rc = install_fba(fd, program_table, stage1b_list,
+				 stage1b_count, info);
 		break;
-	case disk_type_eckd_classic:
-		rc = install_eckd_classic(fd, program_table, stage2_list,
-					  stage2_count, info);
+	case disk_type_eckd_ldl:
+		rc = install_eckd_ldl(fd, program_table, stage1b_list,
+				      stage1b_count, info);
 		break;
-	case disk_type_eckd_compatible:
-		rc = install_eckd_compatible(fd, program_table, info, job);
+	case disk_type_eckd_cdl:
+		rc = install_eckd_cdl(fd, program_table, stage1b_list,
+				      stage1b_count, info);
 		break;
 	case disk_type_diag:
 		/* Should not happen */
@@ -356,7 +366,7 @@ install_bootloader(const char* device, disk_blockptr_t *program_table,
 
 /* Rewind the tape device identified by FD. Return 0 on success, non-zero
  * otherwise. */
-static int
+int
 rewind_tape(int fd)
 {
 	struct mtop op;
@@ -370,7 +380,6 @@ rewind_tape(int fd)
 		return 0;
 }
 
-
 static int
 ask_for_confirmation(const char* fmt, ...)
 {
@@ -627,11 +636,8 @@ overwrite_partition_start(int fd, struct disk_info* info, int mv_dump_magic)
 	const char dump_magic[] = {0xa8, 0x19, 0x01, 0x73,
 		0x61, 0x8f, 0x23, 0xfd};
 
-	if (lseek(fd, info->geo.start * info->phy_block_size, SEEK_SET) !=
-	    (off_t) info->geo.start * info->phy_block_size) {
-		error_text("Could not seek on device");
+	if (misc_seek(fd, info->geo.start * info->phy_block_size))
 		return -1;
-	}
 	if (info->phy_block_size * info->phy_blocks < bytes)
 		bytes = info->phy_block_size * info->phy_blocks;
 	buffer = calloc(1, bytes);
@@ -675,270 +681,241 @@ static int check_eckd_dump_partition(struct disk_info* info)
 	return 0;
 }
 
-struct eckd_dump_param {
-	uint32_t	start_blk;
-	uint32_t	end_blk;
-	uint16_t	blocksize;
-	uint8_t		num_heads;
-	uint8_t		bpt;
-	char		reserved[4];
-} __attribute((packed, may_alias));
-
-static void eckd_dump_store_param(struct eckd_dump_param* param,
-				  struct disk_info* info)
+static void eckd_dump_store_param(struct eckd_dump_param *param,
+				  struct disk_info *info, blocknum_t count)
 {
 	param->start_blk = info->geo.start;
-	param->end_blk = info->geo.start + info->phy_blocks - 1;
+	param->end_blk = info->geo.start + info->phy_blocks - 1 - count;
 	param->num_heads = info->geo.heads;
 	param->blocksize = info->phy_block_size;
-	param->bpt = disk_sec_from_blocknum(param->end_blk, info);
+	param->bpt = info->geo.sectors;
 }
 
-static int
-install_dump_eckd_classic(int fd, struct disk_info* info, uint64_t mem)
+static int install_svdump_eckd_ldl(int fd, struct disk_info *info, uint64_t mem)
 {
-	struct boot_eckd_stage0 stage0;
-	struct boot_eckd_classic_stage1 stage1;
-	void* buffer;
-	size_t size;
-	int rc;
+	disk_blockptr_t *stage2_list, *stage1b_list;
+	blocknum_t stage2_count, stage1b_count;
+	struct boot_eckd_ldl_stage0 stage0;
+	struct boot_eckd_stage1 stage1;
+	struct eckd_dump_param param;
+	size_t stage2_size;
+	void *stage2;
+	int rc = -1;
 
-	/* Prepare stage 0 */
-	boot_init_eckd_classic_stage0(&stage0);
-	/* Install stage 0 */
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage0,
-		sizeof(struct boot_eckd_stage0)));
-	if (rc) {
-		error_text("Could not write boot loader");
-		return rc;
+	if (boot_get_eckd_dump_stage2(&stage2, &stage2_size, mem))
+		goto out;
+	if (blk_cnt(stage2_size, info) > STAGE2_BLK_CNT_MAX) {
+		error_reason("ECKD dump record is too large");
+		goto out_free_stage2;
 	}
-	/* Prepare stage 1 */
-	rc = boot_init_eckd_classic_dump_stage1(&stage1, info);
-	if (rc)
-		return rc;
-	/* Fill in dump partition parameter */
-	eckd_dump_store_param((struct eckd_dump_param *) &stage1.param1, info);
-	/* Install stage 1 */
-	if (lseek(fd, info->phy_block_size, SEEK_SET) !=
-	    info->phy_block_size) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage1,
-		sizeof(struct boot_eckd_classic_stage1)));
-	if (rc) {
-		error_text("Could not write dump record");
-		return rc;
-	}
-	/* Overwrite first 64k of partition */
-	rc = overwrite_partition_start(fd, info, 0);
-	if (rc)
-		return rc;
-	/* Install stage 2 dump record at beginning of partition */
-	if (lseek(fd, info->geo.start * info->phy_block_size, SEEK_SET) !=
-	    (off_t) info->geo.start * info->phy_block_size) {
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = boot_get_eckd_dump_stage2(&buffer, &size, mem);
-	if (rc)
-		return rc;
-	rc = DRY_RUN_FUNC(misc_write(fd, buffer, size));
-	if (rc)
-		error_text("Could not write dump record");
-	free(buffer);
+	if (overwrite_partition_start(fd, info, 0))
+		goto out_free_stage2;
+	/* Install stage 2 and stage 1b to beginning of partition */
+	if (misc_seek(fd, info->geo.start * info->phy_block_size))
+		goto out_free_stage2;
+	stage2_count = disk_write_block_buffer(fd, 1, stage2, stage2_size,
+					       &stage2_list, info);
+	if (stage2_count == 0)
+		goto out_free_stage2_list;
+	if (install_eckd_stage1b(fd, &stage1b_list, &stage1b_count,
+				 stage2_list, stage2_count, info))
+		goto out_free_stage2_list;
+	/* Install stage 0 - afterwards we are at stage 1 position*/
+	boot_init_eckd_ldl_stage0(&stage0);
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage0, sizeof(stage0), 0)))
+		goto out_free_stage1b_list;
+	/* Install stage 1 and fill in dump partition parameter */
+	if (boot_init_eckd_stage1(&stage1, stage1b_list, stage1b_count))
+		goto out_free_stage1b_list;
+
+	eckd_dump_store_param(&param, info, 0);
+	boot_get_dump_info(&stage1.boot_info, BOOT_INFO_DEV_TYPE_ECKD, &param);
+
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage1, sizeof(stage1),
+				     sizeof(stage0))))
+		goto out_free_stage1b_list;
+	rc = 0;
+out_free_stage1b_list:
+	free(stage1b_list);
+out_free_stage2_list:
+	free(stage2_list);
+out_free_stage2:
+	free(stage2);
+out:
 	return rc;
 }
 
-
-static int
-install_dump_eckd_compatible(int fd, struct disk_info* info, uint64_t mem)
+static int install_dump_eckd_cdl(int fd, struct disk_info *info, void *stage2,
+				 size_t stage2_size, int mvdump, int force)
 {
-	struct boot_eckd_stage0 stage0;
-	struct boot_eckd_compatible_stage1 stage1;
-	void* buffer;
-	size_t size;
-	int rc;
+	blocknum_t count, stage2_count, stage1b_count;
+	disk_blockptr_t *stage2_list, *stage1b_list;
+	struct boot_eckd_cdl_stage0 stage0_cdl;
+	struct boot_eckd_stage1 stage1;
+	struct eckd_dump_param param;
+	int rc = -1;
 
-	/* Prepare stage 0 */
-	boot_init_eckd_compatible_stage0(&stage0);
+	count = blk_cnt(stage2_size, info);
+	if (count > STAGE2_BLK_CNT_MAX) {
+		error_reason("ECKD dump record is too large");
+		goto out;
+	}
+	count += blk_cnt(sizeof(struct boot_eckd_stage1b), info);
+	if (count > (blocknum_t) info->geo.sectors - ECKD_CDL_DUMP_REC) {
+		error_reason("ECKD dump record is too large");
+		goto out;
+	}
+	/* Install stage 2 */
+	if (misc_seek(fd, ECKD_CDL_DUMP_REC * info->phy_block_size))
+		goto out;
+	stage2_count = disk_write_block_buffer(fd, 1, stage2, stage2_size,
+					       &stage2_list, info);
+	if (stage2_count == 0)
+		goto out;
+	/* Install stage 1b behind stage 2*/
+	if (install_eckd_stage1b(fd, &stage1b_list, &stage1b_count,
+				 stage2_list, stage2_count, info))
+		goto out_free_stage2_list;
 	/* Install stage 0 */
-	if (lseek(fd, 4, SEEK_SET) != 4) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage0,
-		sizeof(struct boot_eckd_stage0)));
-	if (rc) {
-		error_text("Could not write boot loader");
-		return rc;
-	}
-	/* Prepare stage 1 */
-	rc = boot_init_eckd_compatible_dump_stage1(&stage1, info, 0);
-	if (rc)
-		return rc;
-	/* Fill in dump partition parameter */
-	eckd_dump_store_param((struct eckd_dump_param *) &stage1.param1, info);
-	/* Install stage 1 */
-	if (lseek(fd, info->phy_block_size + 4, SEEK_SET) !=
-	    info->phy_block_size + 4) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage1,
-			sizeof(struct boot_eckd_compatible_stage1)));
-	if (rc) {
-		error_text("Could not write dump record");
-		return rc;
-	}
-	/* Install stage 2 dump record */
-	if (lseek(fd, 3 * info->phy_block_size, SEEK_SET) !=
-	    3 * info->phy_block_size) {
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = boot_get_eckd_dump_stage2(&buffer, &size, mem);
-	if (rc)
-		return rc;
-	rc = DRY_RUN_FUNC(misc_write(fd, buffer, size));
-	if (rc) {
-		error_text("Could not write dump record");
-		return rc;
-	}
-	free(buffer);
-	return overwrite_partition_start(fd, info, 0);
+	boot_init_eckd_cdl_stage0(&stage0_cdl);
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage0_cdl, sizeof(stage0_cdl), 4)))
+		goto out_free_stage1b_list;
+	/* Install stage 1 and fill in dump partition parameter */
+	if (boot_init_eckd_stage1(&stage1, stage1b_list, stage1b_count))
+		goto out_free_stage1b_list;
+
+	eckd_dump_store_param(&param, info, 0);
+	boot_get_dump_info(&stage1.boot_info, BOOT_INFO_DEV_TYPE_ECKD, &param);
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage1, sizeof(stage1),
+				     info->phy_block_size + 4)))
+		goto out_free_stage1b_list;
+	if (!force && overwrite_partition_start(fd, info, mvdump))
+		goto out_free_stage1b_list;
+	rc = 0;
+out_free_stage1b_list:
+	free(stage1b_list);
+out_free_stage2_list:
+	free(stage2_list);
+out:
+	return rc;
 }
 
-
 static int
-install_mvdump_eckd_compatible(int fd, struct disk_info* info, uint64_t mem,
-			       uint8_t force, struct mvdump_parm_table parm)
+install_svdump_eckd_cdl(int fd, struct disk_info *info, uint64_t mem)
 {
-	struct boot_eckd_stage0 stage0;
-	struct boot_eckd_compatible_stage1 stage1;
-	void* buffer;
-	size_t size;
+	size_t stage2_size;
+	void *stage2;
 	int rc;
 
-	/* Prepare stage 0 */
-	boot_init_eckd_compatible_stage0(&stage0);
-	/* Install stage 0 */
-	if (lseek(fd, 4, SEEK_SET) != 4) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
+	if (boot_get_eckd_dump_stage2(&stage2, &stage2_size, mem))
 		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage0,
-		sizeof(struct boot_eckd_stage0)));
-	if (rc) {
-		error_text("Could not write boot loader");
-		return rc;
-	}
-	/* Prepare stage 1 */
-	rc = boot_init_eckd_compatible_dump_stage1(&stage1, info, 1);
-	if (rc)
-		return rc;
-	/* Install stage 1 */
-	if (lseek(fd, info->phy_block_size + 4, SEEK_SET) !=
-	    info->phy_block_size + 4) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage1,
-			sizeof(struct boot_eckd_compatible_stage1)));
-	if (rc) {
-		error_text("Could not write dump record");
-		return rc;
-	}
-	/* Install stage 2 dump record */
-	if (lseek(fd, 3 * info->phy_block_size, SEEK_SET) !=
-	    3 * info->phy_block_size) {
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = boot_get_eckd_mvdump_stage2(&buffer, &size, mem, force, parm);
-	if (rc)
-		return rc;
-	rc = DRY_RUN_FUNC(misc_write(fd, buffer, size));
-	free(buffer);
-	if (rc) {
-		error_text("Could not write dump record");
-		return rc;
-	}
-	return overwrite_partition_start(fd, info, 1);
+	rc = install_dump_eckd_cdl(fd, info, stage2, stage2_size, 0, 0);
+	free(stage2);
+	return rc;
 }
 
+static int
+install_mvdump_eckd_cdl(int fd, struct disk_info *info, uint64_t mem,
+			uint8_t force, struct mvdump_parm_table parm)
+{
+	size_t stage2_size;
+	void *stage2;
+	int rc;
+
+	/* Write stage 2 + parameter block */
+	if (boot_get_eckd_mvdump_stage2(&stage2, &stage2_size, mem,
+					force, parm))
+		return -1;
+	rc = install_dump_eckd_cdl(fd, info, stage2, stage2_size, 1, force);
+	free(stage2);
+	return rc;
+}
+
+int
+install_fba_stage1b(int fd, disk_blockptr_t **stage1b_list,
+		    blocknum_t *stage1b_count, disk_blockptr_t *stage2_list,
+		    blocknum_t stage2_count, struct disk_info *info)
+{
+	struct boot_fba_stage1b *stage1b;
+	int stage1b_size, rc = -1;
+
+	*stage1b_list = NULL;
+	*stage1b_count = 0;
+	stage1b_size = ROUNDUP(sizeof(*stage1b), info->phy_block_size);
+	stage1b = misc_malloc(stage1b_size);
+	if (stage1b == NULL)
+		goto out;
+	memset(stage1b, 0, stage1b_size);
+	if (boot_init_fba_stage1b(stage1b, stage2_list, stage2_count))
+		goto out_free_stage1b;
+	*stage1b_count = disk_write_block_buffer(fd, 1, stage1b, stage1b_size,
+						 stage1b_list, info);
+	if (*stage1b_count == 0)
+		goto out_free_stage1b;
+	rc = 0;
+out_free_stage1b:
+	free(stage1b);
+out:
+	return rc;
+}
 
 static int
-install_dump_fba(int fd, struct disk_info* info, uint64_t mem)
+install_svdump_fba(int fd, struct disk_info *info, uint64_t mem)
 {
+	blocknum_t stage1b_count, stage2_count, blk;
+	disk_blockptr_t *stage1b_list, *stage2_list;
 	struct boot_fba_stage0 stage0;
-	disk_blockptr_t stage2_list[16];
-	blocknum_t block;
-	blocknum_t count;
-	void* buffer;
-	size_t size;
-	int rc;
-	blocknum_t  i;
+	struct fba_dump_param param;
+	size_t stage2_size;
+	void *stage2;
+	int rc = -1;
 
 	/* Overwrite first 64k of partition */
-	rc = overwrite_partition_start(fd, info, 0);
-	if (rc)
-		return rc;
-	/* Install stage 2 dump record at end of partition */
-	rc = boot_get_fba_dump_stage2(&buffer, &size, mem);
-	if (rc)
-		return rc;
-	count = (size + info->phy_block_size - 1) / info->phy_block_size;
-	if (count > 16) {
+	if (overwrite_partition_start(fd, info, 0))
+		goto out;
+	/* Install stage 2 at end of partition */
+	if (boot_get_fba_dump_stage2(&stage2, &stage2_size, mem))
+		goto out;
+	if (blk_cnt(stage2_size, info) > STAGE2_BLK_CNT_MAX) {
 		error_reason("FBA dump record is too large");
-		free(buffer);
-		return -1;
+		goto out_free_stage2;
 	}
-	block = info->geo.start + info->phy_blocks - count;
-	if (lseek(fd, block * info->phy_block_size, SEEK_SET) !=
-	    (off_t) block * info->phy_block_size) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
-		free(buffer);
-		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, buffer, size));
-	free(buffer);
-	if (rc) {
-		error_text("Could not write to device");
-		return rc;
-	}
-	for (i=0; i < count; i++) {
-		stage2_list[i].linear.block = block + i;
-		stage2_list[i].linear.size = info->phy_block_size;
-		stage2_list[i].linear.blockct = 0;
-	}
-	/* Prepare stage 0 */
-	rc = boot_init_fba_stage0(&stage0, stage2_list, count);
-	if (rc)
-		return rc;
-	/* Fill in dump parameters */
-	stage0.param1 = (uint64_t) info->geo.start;
-	stage0.param2 = (uint64_t) block - 1;
-	/* Install stage 0 */
-	if (lseek(fd, 0, SEEK_SET) != 0) {
-		error_reason(strerror(errno));
-		error_text("Could not seek on device");
-		return -1;
-	}
-	rc = DRY_RUN_FUNC(misc_write(fd, &stage0,
-		sizeof(struct boot_fba_stage0)));
-	if (rc)
-		error_text("Could not write dump record");
+	blk = (info->geo.start + info->phy_blocks - blk_cnt(stage2_size, info));
+	if (misc_seek(fd, blk * info->phy_block_size))
+		goto out_free_stage2;
+	stage2_count = disk_write_block_buffer(fd, 1, stage2, stage2_size,
+					       &stage2_list, info);
+	if (stage2_count == 0)
+		goto out_free_stage2;
+	/* Install stage 1b in front of stage 2 */
+	blk -= blk_cnt(sizeof(struct boot_fba_stage1b), info);
+	if (misc_seek(fd, blk * info->phy_block_size))
+		goto out_free_stage2_list;
+	if (install_fba_stage1b(fd, &stage1b_list, &stage1b_count,
+				stage2_list, stage2_count, info))
+		goto out_free_stage2_list;
+	/* Install stage 0/1 fill in dump partition parameter */
+	if (boot_init_fba_stage0(&stage0, stage1b_list, stage1b_count))
+		goto out_free_stage1b_list;
+
+	param.start_blk = (uint64_t) info->geo.start;
+	param.blockct = (uint64_t) blk - 1;
+	boot_get_dump_info(&stage0.boot_info, BOOT_INFO_DEV_TYPE_FBA, &param);
+
+	if (DRY_RUN_FUNC(misc_pwrite(fd, &stage0, sizeof(stage0), 0)))
+		goto out_free_stage1b_list;
+
+	rc = 0;
+out_free_stage1b_list:
+	free(stage1b_list);
+out_free_stage2_list:
+	free(stage2_list);
+out_free_stage2:
+	free(stage2);
+out:
 	return rc;
 }
 
-
 static int
 install_dump_tape(int fd, uint64_t mem)
 {
@@ -966,9 +943,8 @@ install_dump(const char* device, struct job_target_data* target, uint64_t mem)
 	int fd;
 	int rc;
 
-	fd = open(device, O_RDWR);
+	fd = misc_open_exclusive(device);
 	if (fd == -1) {
-		error_reason(strerror(errno));
 		error_text("Could not open dump device '%s'", device);
 		return -1;
 	}
@@ -1029,8 +1005,8 @@ install_dump(const char* device, struct job_target_data* target, uint64_t mem)
 		return -1;
 	}
 	switch (info->type) {
-	case disk_type_eckd_classic:
-	case disk_type_eckd_compatible:
+	case disk_type_eckd_ldl:
+	case disk_type_eckd_cdl:
 		if (check_eckd_dump_partition(info)) {
 			error_text("Dump target '%s'", device);
 			rc = -1;
@@ -1039,8 +1015,8 @@ install_dump(const char* device, struct job_target_data* target, uint64_t mem)
 		/* Fall through. */
 	case disk_type_fba:
 		part_size = info->phy_block_size * info->phy_blocks;
-		printf("Dump target: partition '%s' with a size of "
-		       "%ld MB.\n", device, (long) part_size >> 20);
+		printf("Dump target: partition '%s' with a size of %llu MB.\n",
+		       device, (unsigned long long) part_size >> 20);
 		rc = ask_for_confirmation("Warning: All information on "
 					  "partition '%s' will be lost!\n"
 					  "Do you want to continue creating "
@@ -1051,12 +1027,12 @@ install_dump(const char* device, struct job_target_data* target, uint64_t mem)
 			printf("Installing dump record on partition with %s\n",
 			       disk_get_type_name(info->type));
 		}
-		if (info->type == disk_type_eckd_classic)
-			rc = install_dump_eckd_classic(fd, info, mem);
-		else if (info->type == disk_type_eckd_compatible)
-			rc = install_dump_eckd_compatible(fd, info, mem);
+		if (info->type == disk_type_eckd_ldl)
+			rc = install_svdump_eckd_ldl(fd, info, mem);
+		else if (info->type == disk_type_eckd_cdl)
+			rc = install_svdump_eckd_cdl(fd, info, mem);
 		else
-			rc = install_dump_fba(fd, info, mem);
+			rc = install_svdump_fba(fd, info, mem);
 		break;
 	case disk_type_scsi:
 		error_reason("%s: Unsupported disk type '%s' (try --dumptofs)",
@@ -1095,9 +1071,8 @@ install_mvdump(char* const device[], struct job_target_data* target, int count,
 	parm.num_param = count;
 	parm.timestamp = (time.tv_sec << 20) + time.tv_usec;
 	for (i = 0; i < count; i++) {
-		fd = open(device[i], O_RDWR);
+		fd = misc_open_exclusive(device[i]);
 		if (fd == -1) {
-			error_reason(strerror(errno));
 			error_text("Could not open dump target '%s'",
 				   device[i]);
 			rc = -1;
@@ -1125,7 +1100,7 @@ install_mvdump(char* const device[], struct job_target_data* target, int count,
 			rc = -1;
 			goto out;
 		}
-		if (info[i]->type != disk_type_eckd_compatible) {
+		if (info[i]->type != disk_type_eckd_cdl) {
 			error_reason("Dump target '%s' has to be ECKD DASD "
 				     "with cdl format.", device[i]);
 			rc = -1;
@@ -1153,10 +1128,7 @@ install_mvdump(char* const device[], struct job_target_data* target, int count,
 		parm.param[i].start_blk = info[i]->geo.start;
 		parm.param[i].end_blk = info[i]->geo.start +
 					info[i]->phy_blocks - 1;
-		parm.param[i].end_sec =
-			disk_sec_from_blocknum(info[i]->geo.start +
-					       info[i]->phy_blocks - 1,
-					       info[i]);
+		parm.param[i].bpt = info[i]->geo.sectors;
 		parm.param[i].num_heads = info[i]->geo.heads;
 		parm.param[i].blocksize = info[i]->phy_block_size >> 8;
 		parm.param[i].devno = info[i]->devno;
@@ -1189,7 +1161,7 @@ install_mvdump(char* const device[], struct job_target_data* target, int count,
 			rc = -1;
 			goto out;
 		}
-		fd = open(tempdev, O_RDWR);
+		fd = misc_open_exclusive(tempdev);
 		if (fd == -1) {
 			error_text("Could not open temporary device node '%s'",
 				   tempdev);
@@ -1200,8 +1172,7 @@ install_mvdump(char* const device[], struct job_target_data* target, int count,
 		if (verbose)
 			printf("Installing dump record on target partition "
 			       "'%s'\n", device[i]);
-		rc = install_mvdump_eckd_compatible(fd, info[i], mem, force,
-						    parm);
+		rc = install_mvdump_eckd_cdl(fd, info[i], mem, force, parm);
 		misc_free_temp_dev(tempdev);
 
 		if (fsync(fd))
